from modelx.serialize.jsonvalues import *

_formula = lambda PolicyID, ScenID=1: None

_bases = [
    ".BaseProj",
    ".PV"
]

_allow_none = None

_spaces = [
    "InnerProj",
    "Policy",
    "Assumptions"
]

# ---------------------------------------------------------------------------
# References

pol = ("Interface", (".", "Policy"), "auto")


"""
:mod:`ifrs17sim` IFRS waterfall chart
=====================================

The script draws multiple waterfall charts,
each of which depict sources of changes in balances related to IFRS17.
The script is broken down into pieces of code, and explained in a Jupyter notebook,
:doc:`/projects/notebooks/ifrs17sim/ifrs17sim_charts_baseline`.

The live version of the notebook is available online.

*Launch this notebook online!* |binder ifrs17sim_charts_baseline|

.. include:: /binderlinks.rst
   :start-after: Begin binder ifrs17sim_charts_baseline
   :end-before: End binder ifrs17sim_charts_baseline

"""
from draw_charts import draw_waterfall, get_waterfalldata
import modelx as mx

model = mx.read_model("model")
proj = model.OuterProj[1]

# CSM Amortization

csmrf = get_waterfalldata(
            proj, 
            items=['CSM',
                   'IntAccrCSM',
                   'AdjCSM_FlufCF',
                   'TransServices'],
            length=3,
            reverseitems=['TransServices'])

draw_waterfall(csmrf, title='CSM Amortization')

# Expected Cashflow Rollforwad

estcf = get_waterfalldata(
            proj,
            items=['PV_FutureCF',
                   'EstPremIncome',
                   'EstIntOnCF',
                   'EstAcqCashflow',
                   'EstClaim',
                   'EstExps'],
            length=3,
            reverseitems=['EstPremIncome'])

draw_waterfall(estcf, title='Expected Cashflows')
    
# Actual Cashflow Rollforward
    
actcf = get_waterfalldata(
            proj,
            items=['PremIncome',
                   'IntAccumCF',
                   'ExpsAcqTotal',
                   'BenefitTotal',
                   'ExpsMaintTotal',
                   'ActualNetCF'],
            length=3,
            reverseitems=['ExpsAcqTotal',
                          'BenefitTotal',
                          'ExpsMaintTotal'])

draw_waterfall(actcf, stocks=[0, 5], title='Actual Cashflows')

# IFRS17 Financial Performance

ifrspl = get_waterfalldata(
            proj,
            items=['InsurRevenue',
                   'InsurServiceExps',
                   'InsurFinIncomeExps',
                   'ProfitBefTax'],
            length=5,
            reverseitems=['InsurServiceExps'])

draw_waterfall(ifrspl, stocks=[0, 3], title='IFRS17 Profit/Loss')




import smithwilson as sw
import pandas as pd

# Input - Switzerland EIOPA spot rates with LLP 25 years and extrapolation period of 150 years
# Source: https://wayback.archive-it.org/org-1495/20191229100044/https:/eiopa.europa.eu/Publications/Standards/EIOPA_RFR_20190531.zip
#         EIOPA_RFR_20190531_Term_Structures.xlsx; Tab: RFR_spot_no_VA
rates = [-0.00803, -0.00814, -0.00778, -0.00725, -0.00652,
         -0.00565, -0.0048, -0.00391, -0.00313, -0.00214,
         -0.0014, -0.00067, -0.00008, 0.00051, 0.00108,
         0.00157, 0.00197, 0.00228, 0.0025, 0.00264,
         0.00271, 0.00274, 0.0028, 0.00291, 0.00309]
terms = [float(y + 1) for y in range(len(rates))] # 1.0, 2.0, ..., 25.0
ufr = 0.029
alpha = 0.128562

# Target - Extrapolate to 150 years
terms_ext = [float(y + 1) for y in range(150)] # 1.0, 2.0, ..., 150.0

# Calculate fitted rates based on actual observations and two parametes alpha & UFR
rates_ext = sw.fit_smithwilson_rates(rates_obs=rates, t_obs=terms,
                                     t_target=terms_ext, alpha=alpha, ufr=ufr)

# Display Outputs
# Create dictionary with maturity as key and rate as value
observed = dict(zip(terms, rates))
extrapolated = dict(zip(terms_ext, rates_ext.flatten()))

# Create and print dataframe
print(pd.DataFrame({"Observed": observed, "Extrapolated": extrapolated}))


"""
:mod:`ifrs17sim` CSM waterfall chart
====================================

The script draws a CSM waterfall chart.
The script is broken down into pieces of code, and explained in a Jupyter notebook,
:doc:`/projects/notebooks/ifrs17sim/ifrs17sim_csm_waterfall`.

The live version of the notebook is available online.

*Launch this notebook online!* |binder ifrs17sim_csm_waterfall|

.. include:: /binderlinks.rst
   :start-after: Begin binder ifrs17sim_csm_waterfall
   :end-before: End binder ifrs17sim_csm_waterfall

"""
import modelx as mx
from draw_charts import draw_waterfall, get_waterfalldata

model = mx.read_model("model")
proj = model.OuterProj[1]

csmrf = get_waterfalldata(
        proj,
        items=['CSM',
               'IntAccrCSM',
               'AdjCSM_FlufCF',
               'TransServices'],
        length=15,
        reverseitems=['TransServices'])

draw_waterfall(csmrf)



r"""
Monte Carlo vs Black-Scholes-Merton
===========================================

Time values of options and guarantees for various in-the-moneyness
are calculated using Monte Carlo simulations and the Black-Scholes-Merton
pricing formula for European put options.

The Black-Scholes-Merton pricing formula for European put options
can be expressed as below, where
:math:`X` and :math:`S_{0}` correspond to the sum assured
and the initial account value in this example.

.. math::

    p=Xe^{-rT}N\left(-d_{2}\right)-S_{0}N\left(-d_{1}\right)

    d_{1}=\frac{\ln\left(\frac{S_{0}}{X}\right)+\left(r+\frac{\sigma^{2}}{2}\right)T}{\sigma\sqrt{T}}

    d_{2}=d_{1}-\sigma\sqrt{T}


The graph below shows the results obtained from
the Monte Carlo simulations with 10,000 risk neutral scenarios,
and from the Black-Scholes-Merton formula.

Reference: *Options, Futures, and Other Derivatives* by John C.Hull

.. seealso::

    * :doc:`/libraries/notebooks/savings/savings_example1` notebook in the :mod:`~savings` library


"""
import modelx as mx
import pandas as pd
import matplotlib.pyplot as plt
from scipy.stats import norm, lognorm
import numpy as np

model = mx.read_model("CashValue_ME_EX1")
proj = model.Projection
proj.model_point_table = proj.model_point_moneyness
monte_carlo = pd.Series(proj.pv_claims_over_av('MATURITY'), index=proj.model_point().index)
monte_carlo = list(np.average(monte_carlo[i]) for i in range(1, 10))
S0 = proj.model_point_table['premium_pp'] * proj.model_point_table['policy_count']

    
fig, ax = plt.subplots()
ax.scatter(S0, monte_carlo, s= 10, alpha=1, label='Monte Carlo')
ax.scatter(S0, proj.formula_option_put(120), alpha=0.5, label='Black-Scholes-Merton')
ax.legend()
ax.grid(True)
fig.suptitle('TVOG by ITM')







"""Source module for IFRS17 CSM amortization simulation

This module contains formulas to simulate amortization of
contract service margin (CSM) defined in IFRS17.

This module is a mix-in module to projection module in nestedlife project.
"""

from modelx.serialize.jsonvalues import *

_formula = None

_bases = []

_allow_none = None

_spaces = []

# ---------------------------------------------------------------------------
# Cells

def AccumCF(t):
    """Accumulated cashflows"""
    return -NetInsurAssets(t)


def AcqPremRatio():
    """Ratio of PV Acquisiton Cashflows to PV Premiums.

    The ratio is determined by the expectation at issue.
    """
    pvs = InnerProj(0).PV(0)

    return ((pvs.PV_ExpsCommTotal(0) + pvs.PV_ExpsAcq(0))
            / pvs.PV_PremIncome(0))


def ActualNetCF(t):
    """Acutal net cashflow"""
    return NetInsurCF(t) + IntAccumCF(t)


def AdjCSM_FlufCF(t):
    """Adjustment to CSM for changes in fulfilment cashflows (44(c)->B96-B100)

    Warnings:
        Only B96(b) changes in PV of the future cashflows are implemented.

    TODO: Risk Adjustment is yet to be implemented. At the momement
    this adjustment only considers present value of future cashflows.

    TODO: Loss component for onerous contracts are yet to be implemented.
    At the momemnt this adjustment allows negative CSM.
    """

    return PV_Cashflow(t + 1, t + 1, 0) - PV_Cashflow(t, t + 1, 0)


def AdjLCO_FulfCF(t):
    """Adjustment to Loss Component for changes in fulfilment cashflows"""
    if LossComp(t) > 0:
        return min(LossComp(t), EstClaim(t) + EstExps(t) + RelsRiskAdj(t))
    else:
        return 0


def AmortAcqCashflow(t):
    """Amortization of Acquisition Cash Flows

    Warning:
        Implemented as a constant percentage of actual premiums,
        thus not totalling the original amount if actual != expected.
    """
    return -AcqPremRatio * PremIncome(t)


def AsmpChangeImpact(t):
    """Non-financial assumption changes"""
    return PV_Cashflow(t + 1, t + 1, 0) - PV_Cashflow(t, t + 1, 0)


def CSM(t):
    """floored CSM (38, 44)"""
    if t == 0:
        # Initial recognition (38)
        return max(0, PV_FutureCF(t) - RiskAdjustment(t))
    else:
        # Subsequent recognition (44)
        return max(0, (CSM(t - 1)
                + IntAccrCSM(t - 1)
                + AdjCSM_FlufCF(t - 1)
                - TransServices(t - 1)))


def CSM_Unfloored(t):
    """Unfloored CSM (38, 44)"""
    if t == 0:
        # Initial recognition (38)
        return PV_FutureCF(t) - RiskAdjustment(t)
    else:
        # Subsequent recognition (44)
        return (CSM_Unfloored(t-1)
                + IntAccrCSM(t-1)
                + AdjCSM_FlufCF(t-1)
                - TransServices(t-1))


def CovUnitsBeg1(t):
    """The number of coverage units at `t` after new business"""
    return InsurIF_Beg1(t)


def CovUnitsEnd(t):
    """The number of coverage units at `t`"""
    return InsurIF_End(t)


def EstAcqCashflow(t):
    """Expected Acquisition Cashflow"""

    est = InnerProj(t)
    return (est.ExpsCommInit(t)
            + est.ExpsCommRen(t)
            + est.ExpsAcq(t))


def EstClaim(t):
    """Expected Claims

    Warning:
        Using actuarl invest componets as proxy.
    """
    est = InnerProj(t)
    return est.BenefitTotal(t) - InvstComponent(t)


def EstExps(t):
    """Expected Expenses"""

    est = InnerProj(t)
    return (est.ExpsTotal(t)
            - est.ExpsCommInit(t)
            - est.ExpsCommRen(t)
            - est.ExpsAcq(t))


def EstIntOnCF(t):
    """Expected Interest on future cashflows"""
    return InnerProj(t).PV(t).InterestNetCF(t)


def EstPremIncome(t):
    """Expected Premium Income"""
    return InnerProj(t).PremIncome(t)


def Incr_LossComp(t):
    """Increase in Loss Component"""
    if t == 0:
        return LossComp(t)
    else:
        return max(0, LossComp(t) - LossComp(t - 1))


def IncurClaim(t):
    """Incurred Claims"""
    return BenefitTotal(t) - InvstComponent(t)


def IncurExps(t):
    """Incurred Expenses"""
    return (ExpsTotal(t) - ExpsCommTotal(t) - ExpsAcq(t))


def InsServiceResult(t):
    """Insurance Service Result (80(a), 83-86)"""  
    return InsurRevenue(t) - InsurServiceExps(t)


def InsurFinIncomeExps(t):
    """Insurance Finance Income or Expenses (80(b), 87-92, B128-B136)

    Warning:
        Accounting Policy Choice 88(b) not implemented.
    """
    chg_discrate = (PV_Cashflow(t + 1, t + 1, t + 1)
                    - PV_Cashflow(t + 1, t + 1, t))

    return (EstIntOnCF(t) + chg_discrate 
            - IntAccrCSM(t) + IntAccumCF(t))


def InsurRevenue(t):
    """Insurance Revenue (82-85, B120-B125)"""
    return (EstClaim(t)
            + EstExps(t)
            + RelsRiskAdj(t)
            + TransServices(t)
            + AmortAcqCashflow(t)
            - AdjLCO_FulfCF(t))


def InsurServiceExps(t):
    """Insurance Service Expense (103(b))"""
    return (IncurClaim(t)
            + IncurExps(t)
            + AmortAcqCashflow(t)
            + Incr_LossComp(t)
            - AdjLCO_FulfCF(t))


def IntAccrCSM(t):
    """Interest accreted on CSM (44(b))"""
    return CSM(t) * DiscRate(t, 0)


def InvstComponent(t):
    """Investment Components in Incur Claims

    Warning:
        To be implemented.
    """
    return 0


def LossComp(t):
    """Loss Component"""
    if t == 0:
        # Initial recognition
        return - min(0, PV_FutureCF(t) - RiskAdjustment(t))
    else:
        # Subsequent recognition
        return max(0, LossComp(t - 1) - AdjLCO_FulfCF(t - 1))


def NetBalance(t):
    """Net insurance assets plus accumulated cashflows."""
    return NetInsurAssets(t) + AccumCF(t)


def NetInsurAssets(t):
    """Net Insurance Assets or Liabilities

    Warnings:
        The liabilities for incurred claims are not implemented. 
    """    
    return (PV_FutureCF(t)
            - RiskAdjustment(t)
            - CSM(t))


def PV_Cashflow(t, t_at, t_rate):
    """Present value of future cashflows

    This formula takes 3 time parameters.
    The projection starts from `t`, and the projected cashflows are
    discounted back to `t_at`.
    The discount rates applied are the ones at `t_rate`.

    Args:
        t: Time from which the projection
        t_at: Time discount rates at which are used.
        t_rate: Time to which the cashflows are discounted.
    """
    return InnerProj(t).PV(t_rate).PV_NetCashflow(t_at)


def PV_FutureCF(t):
    """Present value of future cashflows"""
    return PV_Cashflow(t, t, t)


def PV_SumCovUnits(t, t_rate):
    """Present value of cumulatvie coverage units

    The non-economic assumptions used for future estimation are the
    current estimate at time `t`.

    The discount rates used are the ones at time `t_rate`.
    """
    return InnerProj(t).PV(t_rate).PV_SumInsurIF(t)


def ProfitBefTax(t):
    """IFRS Profit before tax"""
    return InsServiceResult(t) + InsurFinIncomeExps(t)


def RelsRiskAdj(t):
    """Release of Risk Adjustment to Revenue

    Warning:
        To be implemented.
    """
    return 0


def RiskAdjustment(t):
    """Risk Adjustment

    Warnings:
        To be implemented
    """
    return 0


def TransServices(t):
    """Transfer of services (44(e)->B119)
    """
    csm_pre_rel = (CSM(t)
                   + IntAccrCSM(t)
                   + AdjCSM_FlufCF(t))

    diff_covunits = CovUnitsBeg1(t) * (1 + InnerProj(0).scen.DiscRate(t))
    pv_sumcovunits_end = PV_SumCovUnits(t + 1, 0)

    return csm_pre_rel * diff_covunits / (diff_covunits + pv_sumcovunits_end)





"""
:mod:`ifrs17sim` CSM waterfall chars
====================================

Draw a graph of CSM amortization pattern.
"""

import numpy as np
import matplotlib.pyplot as plt
import seaborn as sns
sns.set()

# %% Draw Actuarl vs Estimated charts

def _get_est(inner, item, t0, t_max):
    """Get estimated values for ``item`` (nan for t < t0)."""
    cells = inner[t0].cells[item]
    return [cells[t] if t >= t0 else np.nan
            for t in range(t_max)]


def _get_act(outer, item, t0, t_max):
    """Get actual values for ``item`` (nan for t > t0)."""
    cells = outer.cells[item]
    return [cells[t] if t <= t0 else np.nan
            for t in range(t_max)]


def _get_actest(outer, inner, item, t0_max, t_max):
    """Get a pair of actual and estimated values at t0."""
    return (_get_act(outer, item, t0_max, t_max),
            _get_est(inner, item, t0_max, t_max))


def _draw_single_ncf(values, ax, xlim, ls):
    """Draw a plotted line of values."""
    ax.plot(values, marker='o', linestyle=ls)
    ax.set_xlim(right=xlim, left=-1)


def draw_actest(outer, inner, item, act_len, t_max):
    """Draw a graph of actual and estimated"""

    _, axs = plt.subplots(nrows=act_len, sharex=True, sharey=True)

    axs[0].set_title(item)

    for t0 in range(act_len):
        ax = axs[t0]
        act, est = _get_actest(outer, inner, item, t0, t_max + 1)
        _draw_single_ncf(est, ax, t_max, ':')
        _draw_single_ncf(act, ax, t_max, '-')


def draw_actest_pairs(outer, inner, items, act_len, t_max):
    """Draw pairs of line graphs."""
    ncols = len(items)
    nrows = act_len

    if ncols == 1:
        return draw_actest(outer, inner, items[0], act_len, t_max)

    _, axs = plt.subplots(nrows=nrows, ncols=ncols, sharex=True)

    for col, item in enumerate(items):
        axs[0][col].set_title(items[col])
        for t0 in range(nrows):
            ax = axs[t0][col]
            act, est = _get_actest(outer, inner, item, t0, t_max + 1)
            _draw_single_ncf(est, ax, t_max, ':')
            _draw_single_ncf(act, ax, t_max, '-')

# %% Waterfall chart functions

def get_waterfalldata(space, items, length, reverseitems=[]):
    """Create a DataFrame for drawing waterfall chart"""

    data = type(space.cells)(
        space.cells, space.cells.impl, keys=items).to_frame(range(length))
    for outflow in reverseitems:
        data[outflow] = -1 * data[outflow]
    return data


def draw_waterfall(df, ax=None, stocks=[0], **kwargs):
    """Draw waterfall chart

    `stocks` parameter is for specifying column indexes of `df` that are to
    be interpreted as stock items as opposed to flow items,
    i.e. bars for these columns are drawn from zero.
    """

    if ax is None:
        plt.figure()
        ax = plt.gca()

    cols = len(df.columns)
    cumsum = df.copy()
    bottom = df.copy()
    bottom.iloc[:,:] = 0

    for c in range(cols):
        if (c + 1) % cols in stocks:
            cumsum.iloc[:, list(range(c + 1))] = 0
            bottom.iloc[:, c] = 0
        else:
            bottom.iloc[:, c] = cumsum.cumsum(axis=1).iloc[:, c]

    data = df.stack()
    bottom = bottom.stack().shift(1, axis=0).fillna(0)

    palette = WaterfallColorPalette(df)
    xlabel = [idx[1] + '(' + str(idx[0]) + ')' for idx in df.stack().index]
    ax = sns.barplot(data=[[i] for i in data], bottom=list(bottom),
                     palette=palette.colors,
                     ax=ax)
    ax.set_xticklabels(labels=xlabel, rotation='vertical')
    if 'title' in kwargs:
        ax.set_title(kwargs['title'])
    ax.get_figure().tight_layout()
    return ax


# %% Unused

class WaterfallColorPalette:
    """Returns color palette for drawing waterfall bars.

    The length of `data` columns

    Args:
        data: a DataFrame of waterfall data.
        length: the number of bars in an interval. Defaults to the length of
                `data` columns.
    """
    
    def __init__(self, data, length=None):

        self.data = data
        if length is None:
            self.length = len(data.columns)
        else:
            self.length = length

        self.paired = plt.get_cmap('Paired').colors
        self.colors = self.set_colors()

    def set_colors(self):

        result = []
        row_count, _ = self.data.shape

        for row in range(row_count):
            for col in range(self.length):
                sign = self.data.iloc[row, col] >= 0
                result.append(self.get_color(col, sign))

        return result

    def get_color(self, idx, sign):
        pair_count = len(self.paired) // 2
        idx = 2 * (idx % pair_count)
        return self.paired[idx + 1] if sign else self.paired[idx]


class CustomColorMap:
    """Unused. Colors changing by size and sign of data

    Args:
        data: a sequence of bar values.
    """
    import math

    def __init__(self, data):

        self.pallen = 20
        self.palette = pal = sns.color_palette("RdBu", self.pallen)

        self.posmax = max([val for val in data if val >= 0] or [0])
        self.negmin = min([val for val in data if val < 0] or [0])

    def get_index(self, value):

        if value > self.posmax:
            return self.pallen - 1
        elif value < self.negmin:
            return 0
        else:
            absmax = self.posmax if value >= 0 else abs(self.negmin)

            idx0 = self.pallen / 2
            return min(idx0 + math.floor(idx0 * value / absmax),
                       self.pallen - 1)

    def get_colors(self, data):
        idxs = [int(self.get_index(val)) for val in data]
        return [self.palette[i] for i in idxs]

# %% Draw stacked bar pair charts

def draw_stackedbarpairs(data, ax=None, **kwargs):
    """Draw pairs of stacked bars

    Draw a series of pairs of stacked bars. For each column,
    bars are drawn either on the left or right of the x ticks,
    depending the sines of values (left if negative, right if positive).
    Bars on each side of each x tick are stacked each other.

    Args:
        data: DataFrame
    """

    if ax is None:
        ax = plt.gca()

    rbars = data[data >= 0]
    lbars = -1 * data[data < 0]

    rbottom, lbottom = \
        [bars.fillna(0).cumsum(axis=1).shift(1, axis=1).fillna(0)
         for bars in [rbars, lbars]]

    def draw_single_bar(data, bottom, ax, n, **kwargs):
        width = 0.4
        return ax.bar(np.arange(len(data)) + n * (width / 2), 
                      data, width - 0.05, bottom=bottom, **kwargs)
    
    legend = [[],[]]
    for c in data.columns:
        bar = draw_single_bar(rbars[c], rbottom[c], ax, +1, label=c)
        color = bar.patches[0].get_facecolor()
        draw_single_bar(lbars[c], lbottom[c], ax, -1, color=color)
        # legend[0].append(bar)
        # legend[1].append(c)
    
    ax.legend()

    if 'title' in kwargs:
        ax.set_title(kwargs['title'])



"""Module to create ``Input`` space and its child spaces from input data.

This module contains a function :py:func:`build_input`,
which creates "Input" space and its subspaces by reading
input data into them from an Excel file.

.. figure:: /images/input_tree.png
    :width: 50%

By default, :py:func:`build_input` reads data from *input.xlsm* in the
same folder as this module, and the structure of ``Input`` space is illustrated
by the diagram below.


.. figure:: /images/input_drawio.png
   :width: 50%

Below are descriptions of Input spaces and its subspaces.

Input
    The parent module of all other input spaces.

    Some tables in *input.xlsm* are imported as cells directly under this space.



PolicyData
    The sample policy data read from *PolicyData* tab in *input.xlsm*.
    Records and fields of *PolicyData* represents policies and policy
    attributes respectively.
    Each record of *PolicyData* is accessible as a dynamic space under
    ``PolicyData`` space indexed by *Policy* column in the input table.
    Each field of a record is accessible as a scalar cells in the dynamic
    space corresponding to the record.

    For example::

        >>> model.Input.PolicyData[3].PolicyTerm()

    gives the value of the policy term of the policy record 3.

MortalityTables
    The sample mortality tables read from *Mortality* tab.
    For Each mortality table, a dynamic space is created indexed
    by *TableID*, and `MortalityTable` cells indexed
    by ``Sex`` and ``Age`` is created in that space.

    The script below will return the mortality rate of Male, Age 41
    from the mortality table 3::

        >>> model.Input.MortalityTables[3].MortalityTable('M', 41)


ProductSpec
    The sample product spec table read from *ProductSpec* tab.
    ``ProductSpec`` dynamic subspaces holds parameters to specify products
    as cells. The dynamic subspaces are indexed by ``Product``,
    ``PolType`` and ``Gen``, and
    each of the subspace has scalar cells defined by the columns
    of *ProductSpec* tab other than the index columns.

    Empty cells in the index columns are imported as ``None``.
    ``Cells.match`` method treats ``None`` as the wildcard when
    finding the closest matching indexes for the given arguments.


Assumption
    The sample assumption table read from *Assumption* tab.

    The ``Assumption`` space has dynamic subspaces indexed by ``Product``,
    ``PolType`` and ``Gen``, each of which has scalar cells defined by the
    columns of *Assumption* tab other than the index columns.

    Empty cells in the index columns are imported as ``None``.
    ``Cells.match`` method treats ``None`` as the wildcard when
    finding the closest matching indexes for the given arguments.

AssumptionTables
    The sample assumption table read from *AssumptionTables* tab.
    The tab holds assumptions by policy duration, such as mortality factors
    and lapse assumptions.

    Each assumption in the table is imported as a cells in the
    ``AssumptionTables`` space.

Scenarios
    The sample scenario table read from *Scenarios* tab.
    TODO


"""
import sys
import os.path as path
import time
from textwrap import dedent

default_input = path.join(path.abspath(path.dirname(__file__)), 'input.xlsx')


class _PrintElapsedTime:

    def __init__(self):
        self.last_time = time.time()

    def set_start(self, msg):
        print(msg, end='', file=sys.stderr)
        self.last_time = time.time()

    def print_time(self, msg):
        this_time = time.time()
        print(msg, "({:.2f}secs)".format(this_time - self.last_time),
              file=sys.stderr)
        self.last_time = this_time


def build_input(model, input_file=default_input):
    """Create "Input" space and its sub spaces by reading data from an Excel
    file into it.

    This function creates a space named "Input" under ``model``,
    reads tables from ``input_file`` into cells and subspaces
    under the "Input" space.

    By default, this function assumes reading data from *input.xlsm*
    located in the same folder as this module.

    Args:
        model: Model object in which 'Input' space is created.
        input_file(str): Path to the Excel input file.
    """

    inp = model.new_space(name='Input')
    inp.allow_none = True
    print_time = True

    print("Started loading data from '%s'." % input_file, file=sys.stderr)

    if print_time:
        timestamp = _PrintElapsedTime()
        timestamp.set_start('Loading PolicyData...')

    policydata = inp.new_space_from_excel(
        book=input_file,
        range_='B7:O307',
        sheet='PolicyData',
        name='PolicyData',
        names_row=0,
        param_cols=[0],
        space_param_order=[0],
        cells_param_order=[])

    if print_time:
        timestamp.print_time('Done.')
        timestamp.set_start('Loading MortalityTables...')

    mortbl = inp.new_space_from_excel(
        book=input_file,
        range_='E4:Q137',
        sheet='Mortality',
        name='MortalityTables',
        names_row=0,
        param_cols=[0],
        names_col=0,
        param_rows=[1, 2],
        space_param_order=[1],
        cells_param_order=[2, 0])

    if print_time:
        timestamp.print_time('Done.')
        timestamp.set_start('Loading ProductSpec...')

    inp.new_space_from_excel(
        book=input_file,
        range_='B7:V22',
        sheet='ProductSpec',
        name='ProductSpec',
        names_row=0,
        param_cols=[0, 1, 2],
        cells_param_order=[0, 1, 2])

    if print_time:
        timestamp.print_time('Done.')
        timestamp.set_start('Loading OtherParam1...')

    inp.new_cells_from_excel(
        book=input_file,
        range_='C6:D9',
        sheet='OtherParams',
        names_row=0,
        param_cols=[0],
        param_order=[0])

    if print_time:
        timestamp.print_time('Done.')
        timestamp.set_start('Loading OtherParams2...')

    inp.new_cells_from_excel(
        book=input_file,
        range_='G6:H9',
        sheet='OtherParams',
        names_row=0,
        param_cols=[0],
        param_order=[0])

    if print_time:
        timestamp.print_time('Done.')
        timestamp.set_start('Loading Assumption...')

    inp.new_space_from_excel(
        book=input_file,
        range_='B7:T17',
        sheet='Assumption',
        name='Assumption',
        names_row=0,
        param_cols=[0, 1, 2],
        cells_param_order=[0, 1, 2])

    if print_time:
        timestamp.print_time('Done.')
        timestamp.set_start('Loading AssumptionTables...')

    inp.new_space_from_excel(
        book=input_file,
        range_='B7:J27',
        sheet='AssumptionTables',
        name='AssumptionTables',
        names_row=0,
        param_cols=[0],
        cells_param_order=[0])

    if print_time:
        timestamp.print_time('Done.')
        timestamp.set_start('Loading Scenarios...')

    scenarios = inp.new_space_from_excel(
        book=input_file,
        range_='A3:C1513',
        sheet='Scenarios',
        name='Scenarios',
        names_row=0,
        param_cols=[0, 1],
        space_param_order=[0],
        cells_param_order=[1])

    if print_time:
        timestamp.print_time('Done.')

    print(dedent("""\
        Input space and its sub spaces are saved in '[project name].mx'.
        You can load input data from the saved file instead of 'input.xlsx'
        by passing 'load_saved=True' to simplelife.build function."""),
          file=sys.stderr)

    return inp



"""Source module for Life SCR standard formulas

This module contains formulas to calculate SCR.

.. rubric:: Common cells parameters


Attributes:
    risk: one of the following strings that represents a life sub-module.
        only Mortality, Longevity, Lapse and Life expense risks are
        implemented.

* "mort": Mortality risk
* "longev": Longevity risk
* "disab": Disability risk
* "lapse": Lapse risk
* "exps": Life expense risk
* "rev": Revision risk
* "cat": Life catastrophe risk

"""

def SCR_life():
    """The capital requirement for life underwriting risk

    .. math::
        \sqrt{\sum_{i,j}Corr_{(i,j)}\cdot SCR_i\cdot SCR_j}

    where i, j are a combination of risks

    """
    return sum(Life(r) * Life(c) * Corr(r, c) for r, c in Corr) ** 0.5


def Life(risk):
    """The capital requirement for each risk under the life underwriting risk

    Args:
        risk: string that represents life risk sub-module ("mort",
            "longev", "disab", "lapse", "exps", "rev", "cat")
    """

    if risk == 'lapse':
        return max(LapseRisk(shock) for shock in ['up', 'down', 'mass'])    
    else:
        return max(NetAstValue() - NetAstValue(risk), 0)


def LapseRisk(shock):
    """The capital requirement for lapse risk for each shock

    Args:
        shock: string that represents each lapse shock
            ("up", "down", "mass")
    """
    
    if shock in ['up', 'down']:
        return max(NetAstValue() - NetAstValue('lapse', shock), 0)
    
    elif shock == 'mass':
        
        retail_share = 1
        nonretail_share = 1 - retail_share
        
        retail = NetAstValue() - NetAstValue('lapse', shock) \
            if retail_share else 0
                
        nonretail = NetAstValue() - NetAstValue('lapse', shock, 'noretail') \
            if nonretail_share else 0
        
        return max(retail_share * retail + nonretail_share * nonretail, 0)

    else:
        raise ValueError("Unknown shock: %s" % shock)


def NetAstValue(risk='base', shock=None, scope=None):
    """Net value of assets minus liabilities

    This formula is simplified and present value of net
    liability cashflows is used in replacement for net asset value,
    based on the assumption that the value of assets
    does not change by life risk scenarios.

    Args:
        risk: string that represents life risk sub-module ("mort",
            "longev", "disab", "lapse", "exps", "rev", "cat")
        shock: string that represents each lapse shock
            ("up", "down", "mass")
        scope: "nonretail" or None (by default)

    """
    return Projection[risk, shock, scope].PV_NetCashflow(t0)



"""
:mod:`nestedlife` cashflow
==========================

Draw a graph of present value of cashflow
"""
import modelx as mx

model = mx.read_model("model")

# Policy point ID and aliases
polid = 171
outer = model.OuterProj

# Code block for overwiting the defaut model

outer[polid].asmp.SurrRateMult[1] = 2
outer[polid].asmp.SurrRateMult[2] = 0.5
outer[polid].asmp.SurrRateMult[3] = 1

outer[polid].InnerProj[1].asmp.SurrRateMult[1] = 2
outer[polid].InnerProj[2].asmp.SurrRateMult[2] = 0.5
outer[polid].InnerProj[3].asmp.SurrRateMult[3] = 1

# Code block for PV graph

import numpy as np
import matplotlib.pyplot as plt
import seaborn as sns
sns.set()    
    
    
def draw_bars(item):
    
    term = 5
    
    expect = []
    for t0 in range(term):
        expect_t0 = [np.nan] * term
        for t in range(t0, term):
            cells = outer[polid].InnerProj[t0].cells[item]
            expect_t0[t] = cells[t]
            
        expect.append(expect_t0)
    
    fg, ax = plt.subplots()
    ax.set_xlim(left=-0.5, right=term + 1)
    
    for t0 in range(term):
        draw_single_bar(expect[t0], ax, t0)
    
    
def draw_single_bar(data, ax, t0):

    size = len(data)
    width = 1/ (size + 1)
    ax.bar(np.arange(size) + t0 * (width + 0.05), data, width)
    

# PV Test
if __name__ == '__main__':
    draw_bars('PV_NetCashflow')




"""Policy data and attributes

This Space contains policy data and Cells associated with
policy attributes and policy values.

.. rubric:: References

Attributes:
    PolicyData: `PandasData`_ object holding model point data as a pandas
        DataFrame read from the model point file.
    ProductSpec: `ExcelRange`_ object which is a dict like object for associating
        product spec parameters to model point keys.
    LifeTable: Alias for :mod:`~fastlife.model.LifeTable`
    PremTerm: Alias for :func:`PolicyTerm`


.. _PandasData:
   https://docs.modelx.io/en/latest/reference/dataclient.html#pandasdata

.. _ExcelRange:
   https://docs.modelx.io/en/latest/reference/dataclient.html#excelrange

"""

from modelx.serialize.jsonvalues import *

_formula = None

_bases = []

_allow_none = True

_spaces = []

# ---------------------------------------------------------------------------
# Cells

def AnnPremRate():
    """Annualized Premium Rate per Sum Assured"""
    return GrossPremRate() * PremFreq().mask(PremFreq() == 0, other=1/10)


def CashValueRate(t):
    """Cash Value Rate per Sum Assured"""
    return np.maximum(ReserveNLP_Rate('PREM', t) - SurrCharge(t), 0)


def GrossPremRate():
    """Gross Premium Rate per Sum Assured per payment"""

    def get_value(pol):

        prod = pol['Product']
        alpha = pol['LoadAcqSA']
        beta = pol['LoadMaintPrem']
        delta = pol['LoadMaintPrem2']
        gamma = pol['LoadMaintSA']
        gamma2 = pol['LoadMaintSA2']
        freq = pol['PremFreq']

        x, n, m = pol['IssueAge'], pol['PolicyTerm'], pol['PolicyTerm']

        comf = LifeTable[pol['Sex'], pol['IntRate_PREM'], pol['TableID_PREM']]

        if prod == 'TERM' or prod == 'WL':
            return (comf.Axn(x, n) + alpha + gamma * comf.AnnDuenx(x, n, freq)
                    + gamma2 * comf.AnnDuenx(x, n-m, 1, m)) / (1-beta-delta) / freq / comf.AnnDuenx(x, m, freq)

        elif prod == 'ENDW':
            return (comf.Exn(x, n) + comf.Axn(x, n) + alpha + gamma * comf.AnnDuenx(x, n, freq)
                    + gamma2 * comf.AnnDuenx(x, n-m, 1, m)) / (1-beta-delta) / freq / comf.AnnDuenx(x, m, freq)
        else:
            raise ValueError('invalid product')


    result = PolicyDataExt1().apply(get_value, axis=1)
    result.name = 'GrossPremRate'

    return result


def GrossPremTable():
    """Gross premium table"""
    return None


def InitSurrCharge():
    """Initial Surrender Charge Rate"""

    def get_value(pol):

        prod, polt, gen = pol['Product'], pol['PolicyType'], pol['Gen']
        term = pol['PolicyTerm']

        param1 = SpecLookup.match("SurrChargeParam1", prod, polt, gen).value
        param2 = SpecLookup.match("SurrChargeParam2", prod, polt, gen).value

        if param1 is None or param2 is None:
            raise ValueError('SurrChargeParam not found')

        return param1 + param2 * min(term / 10, 1)


    result = PolicyData().apply(get_value, axis=1)
    result.name = 'InitSurrCharge'

    return result


def IntRate(RateBasis):
    """Interest Rate"""

    if RateBasis == 'PREM':
        basis = 'IntRatePrem'
    elif RateBasis == 'VAL':
        basis = 'IntRateVal'
    else:
        raise ValueError('invalid RateBasis')


    def get_value(pol):

        result = SpecLookup.match(basis, 
                                  pol["Product"], 
                                  pol["PolicyType"],
                                  pol["Gen"]).value

        if result is not None:
            return result
        else:
            raise ValueError('lookup failed')


    result = PolicyData().apply(get_value, axis=1)
    result.name = 'IntRate_' + RateBasis
    return result


def LoadAcqSA():
    """Acquisition Loading per Sum Assured"""
    param1 = Product().apply(lambda prod: SpecLookup("LoadAcqSAParam1", prod))
    param2 = Product().apply(lambda prod: SpecLookup("LoadAcqSAParam2", prod))

    result = param1 + param2 * np.minimum(PolicyTerm() / 10, 1)

    result.name = 'LoadAcqSA'
    return result


def LoadMaintPrem():
    """Maintenance Loading per Gross Premium"""

    def get_value(pol):

        if SpecLookup("LoadMaintPremParam1", pol["Product"]) is not None:
            return SpecLookup("LoadMaintPremParam1", pol["Product"])

        elif SpecLookup("LoadMaintPremParam2", pol["Product"]) is not None:
            param = SpecLookup("LoadMaintPremParam2", pol["Product"])
            return (param + min(10, pol["PolicyTerm"])) / 100

        else:
            raise ValueError('LoadMaintPrem parameters not found')


    result = PolicyData().apply(get_value, axis=1)
    result.name = 'LoadMaintPrem'

    return result


def LoadMaintSA():
    """Maintenance Loading per Sum Assured during Premium Payment"""

    def get_value(pol):

        result = SpecLookup.match("LoadMaintSA", 
                                  pol["Product"], 
                                  pol["PolicyType"],
                                  pol["Gen"]).value

        if result is not None:
            return result
        else:
            raise ValueError('lookup failed')


    result = PolicyData().apply(get_value, axis=1)
    result.name = 'LoadMaintSA'
    return result


def LoadMaintSA2():
    """Maintenance Loading per Sum Assured after Premium Payment"""

    def get_value(pol):

        result = SpecLookup.match("LoadMaintSA2", 
                                  pol["Product"], 
                                  pol["PolicyType"],
                                  pol["Gen"]).value

        if result is not None:
            return result
        else:
            raise ValueError('lookup failed')


    result = PolicyData().apply(get_value, axis=1)
    result.name = 'LoadMaintSA2'
    return result


def NetPremRate(basis):
    """Net Premium Rate"""

    def get_value(pol):

        prod = pol['Product']
        gamma2 = pol['LoadMaintSA2']

        x, n, m = pol['IssueAge'], pol['PolicyTerm'], pol['PolicyTerm']

        comf = LifeTable[pol['Sex'], pol['IntRate_' + basis], pol['TableID_' + basis]]

        if prod == 'TERM' or prod == 'WL':
            return (comf.Axn(x, n) + gamma2 * comf.AnnDuenx(x, n-m, 1, m)) / comf.AnnDuenx(x, n)

        elif prod == 'ENDW':
            return (comf.Axn(x, n) + gamma2 * comf.AnnDuenx(x, n-m, 1, m)) / comf.AnnDuenx(x, n)

        else:
            raise ValueError('invalid product')


    result = PolicyDataExt1().apply(get_value, axis=1)
    result.name = 'NetPremRate_' + basis

    return result


def ReserveNLP_Rate(basis, t):
    """Net level premium reserve rate"""

    def get_value(pol):

        prod = pol['Product']
        gamma2 = pol['LoadMaintSA2']
        netp = pol['NetPremRate_' + basis]

        x, n, m = pol['IssueAge'], pol['PolicyTerm'], pol['PolicyTerm']

        lt = LifeTable[pol['Sex'], pol['IntRate_' + basis], pol['TableID_' + basis]]

        if t <= m:
            return lt.Axn(x+t, n-t) + (gamma2 * lt.AnnDuenx(x+t, n-m, 1, m-t)
                    - netp * lt.AnnDuenx(x+t, m-t))
        elif t <=n:
            return lt.Axn(x+t, n-t) + gamma2 * lt.AnnDuenx(x+t, n-m, 1, m-t)
        else:
            return 0


    result = PolicyDataExt2().apply(get_value, axis=1)
    result.name = 'ReserveNLP_Rate'

    return result


def ReserveRate():
    """Valuation Reserve Rate per Sum Assured"""
    return None


def SurrCharge(t):
    """Surrender Charge Rate per Sum Assured"""
    m = PremTerm()
    return InitSurrCharge * np.maximum((np.minimum(m, 10) - t) / np.minimum(m, 10), 0)


def TableID(RateBasis):
    """Mortality Table ID"""

    if RateBasis == 'PREM':
        basis = "MortTablePrem"
    elif RateBasis == 'VAL':
        basis = "MortTableVal"
    else:
        raise ValueError('invalid RateBasis')


    def get_value(pol):

        result = SpecLookup.match(basis, 
                                  pol["Product"], 
                                  pol["PolicyType"],
                                  pol["Gen"]).value

        if result is not None:
            return result
        else:
            raise ValueError('lookup failed')


    result = PolicyData().apply(get_value, axis=1)
    result.name = 'TableID_' + RateBasis
    return result


def UernPremRate():
    """Unearned Premium Rate"""
    return None


Product = lambda: PolicyData()['Product']

PolicyType = lambda: PolicyData()['PolicyType']

Gen = lambda: PolicyData()['Gen']

Channel = lambda: PolicyData()['Channel']

Sex = lambda: PolicyData()['Sex']

Duration = lambda: PolicyData()['Duration']

IssueAge = lambda: PolicyData()['IssueAge']

PremFreq = lambda: PolicyData()['PremFreq']

PolicyTerm = lambda: PolicyData()['PolicyTerm']

PolicyCount = lambda: PolicyData()['PolicyCount']

SumAssured = lambda: PolicyData()['SumAssured']

def LoadMaintPrem2():
    """Maintenance Loading per Gross Premium for Premium Waiver"""

    result = pd.Series(0.002, index=PolicyData().index)

    result[PremTerm < 10] = 0.001
    result[PremTerm < 5] = 0.0005

    result.name = 'LoadMaintPrem2'

    return result


def PolicyDataExt1():
    """Extended Poicy Data"""

    data = pd.concat([PolicyData(), 
                      LoadAcqSA(),
                      LoadMaintPrem(),
                      LoadMaintPrem2(),
                      LoadMaintSA(),
                      LoadMaintSA2(),
                      IntRate('PREM'),
                      TableID('PREM')], axis=1)

    return data


def PolicyDataExt2():
    """Extended Poicy Data"""

    data = pd.concat([PolicyDataExt1(),
                      NetPremRate('PREM')], axis=1)

    return data


def SpecLookup(spec, prod=None, polt=None, gen=None):
    """Look up product specs"""
    return ProductSpec.get((spec, prod, polt, gen), None)


# ---------------------------------------------------------------------------
# References

LifeTable = ("Interface", ("...", "LifeTable"), "auto")

PremTerm = ("Interface", (".", "PolicyTerm"), "auto")

ProductSpec = ("Pickle", 3020542092744)

PolicyData = ("Pickle", 3020548337160)


"""
:mod:`ifrs17sim` IFRS balancesheet itms
=======================================

Fulfilment CF, CSM, Cash balances
"""

# When the current directory is this folder,
# The try-except statement below can be replaced by just the last two
# import statements.
import draw_charts
import modelx as mx

model = mx.read_model("model")
proj = model.OuterProj[171]


ifrsbs = proj.cells['CSM',
                    'PV_FutureCF',
                    'AccumCF'].to_frame(range(10))

ifrsbs.columns = ['CSM', 'FCF', 'Cash']
ifrsbs['FCF'] = -1 * ifrsbs['FCF']
ifrsbs['Cash'] = -1 * ifrsbs['Cash']

draw_charts.draw_stackedbarpairs(ifrsbs,
                                 title='Fulfilment CF and CSM')




"""Source module to create ``Policy`` space from.

This module is a source module to create ``Policy`` space and its
sub spaces from.
The formulas of the cells in the ``Policy`` space are created from the
functions defined in this module.

The ``Policy`` space is the base space of the policy spaces
for individual policies, which are derived from and belong to
the ``Policy`` space as its dynamic child spaces.

The policy spaces for individual policies are parametrized by ``PolicyID``.
For example, to get the policy space of the policy whose ID is 171::

    >> pol = model.Policy(171)

The cells in a policy space for each individual policy retrieve
input data, calculate and hold values of policy attributes specific to that policy,
so various spaces in :mod:`Input<simplelife.build_input>` must be accessible
from the ``Policy`` space.

.. rubric:: Project Templates

This module is included in the following project templates.

* :mod:`simplelife`
* :mod:`nestedlife`

.. rubric:: Space Parameters

Attributes:
    PolicyID: Policy ID

.. rubric:: References in Base

Attributes:
    PolicyData: Input.PolicyData
    ProductSpec: Input.ProductSpec
    LifeTable: LifeTable

.. rubric:: References in Sub

Attributes:
    Product: Product key
    PolicyType: Policy type key
    Gen: Generation key
    Channel: Channel key
    Sex: ``M`` for Male, ``F`` for Female
    Duration: Number of years lapsed. 0 for new business
    IssueAge: Issue age
    PremFreq: Number of premium payments per year. 12 for monthly payments
    PolicyTerm: Policy term in year
    PolicyCount: Number of policies
    SumAssured: Sum Assured per policy
"""

from modelx.serialize.jsonvalues import *

_formula = lambda PolicyID: None

_bases = []

_allow_none = None

_spaces = []

# ---------------------------------------------------------------------------
# Cells

def AnnPremRate():
    """Annualized Premium Rate per Sum Assured"""
    return GrossPremRate * (1/10 if PremFreq == 0 else PremFreq)


def CashValueRate(t):
    """Cash Value Rate per Sum Assured"""
    return max(ReserveNLP_Rate('PREM', t) - SurrCharge(t), 0)


def GrossPremRate():
    """Gross Premium Rate per Sum Assured per payment"""

    alpha = LoadAcqSA()
    beta = LoadMaintPrem()
    gamma = LoadMaintSA()
    gamma2 = LoadMaintSA2()
    delta = LoadMaintPremWaiverPrem()

    x, n, m = IssueAge(), PolicyTerm(), PremTerm()

    comf = LifeTable[Sex(), IntRate('PREM'), TableID('PREM')]

    if Product == 'TERM' or Product == 'WL':
        return (comf.Axn(x, n) + alpha + gamma * comf.AnnDuenx(x, n, PremFreq)
                + gamma2 * comf.AnnDuenx(x, n-m, 1, m)) / (1-beta-delta) / PremFreq / comf.AnnDuenx(x, m, PremFreq)

    elif Product == 'ENDW':
        return (comf.Exn(x, n) + comf.Axn(x, n) + alpha + gamma * comf.AnnDuenx(x, n, PremFreq)
                + gamma2 * comf.AnnDuenx(x, n-m, 1, m)) / (1-beta-delta) / PremFreq / comf.AnnDuenx(x, m, PremFreq)
    else:
        raise ValueError('invalid product')


def GrossPremTable():
    """Gross premium table"""
    return None


def InitSurrCharge():
    """Initial Surrender Charge Rate"""

    param1 = SpecLookup.match("SurrChargeParam1", Product(), PolicyType(), Gen()).value
    param2 = SpecLookup.match("SurrChargeParam2", Product(), PolicyType(), Gen()).value

    if param1 is None or param2 is None:
        raise ValueError('SurrChargeParam not found')

    return param1 + param2 * min(PolicyTerm / 10, 1)


def IntRate(RateBasis):
    """Interest Rate"""

    if RateBasis == 'PREM':
        basis = 'IntRatePrem'
    elif RateBasis == 'VAL':
        basis = 'IntRateVal'
    else:
        raise ValueError('invalid RateBasis')

    result = SpecLookup.match(basis, Product(), PolicyType(), Gen()).value

    if result is not None:
        return result
    else:
        raise ValueError('invalid RateBais')


def LoadAcqSA():
    """Acquisition Loading per Sum Assured"""
    param1 = SpecLookup("LoadAcqSAParam1", Product())
    param2 = SpecLookup("LoadAcqSAParam2", Product())

    return param1 + param2 * min(PolicyTerm / 10, 1)


def LoadMaintPrem():
    """Maintenance Loading per Gross Premium"""

    if SpecLookup("LoadMaintPremParam1", Product()) is not None:
        return SpecLookup("LoadMaintPremParam1", Product())

    elif SpecLookup("LoadMaintPremParam2", Product()) is not None:
        param = SpecLookup("LoadMaintPremParam2", Product())
        return (param + min(10, PolicyTerm)) / 100

    else:
        raise ValueError('LoadMaintPrem parameters not found')


def LoadMaintPremWaiverPrem():
    """Maintenance Loading per Gross Premium for Premium Waiver"""

    if PremTerm < 5:
        return 0.0005
    elif PremTerm < 10:
        return 0.001
    else:
        return 0.002


def LoadMaintSA():
    """Maintenance Loading per Sum Assured during Premium Payment"""

    result = SpecLookup.match("LoadMaintSA", Product(), PolicyType(), Gen()).value

    if result is not None:
        return result
    else:
        raise ValueError('lookup failed')


def LoadMaintSA2():
    """Maintenance Loading per Sum Assured after Premium Payment"""

    result = SpecLookup.match("LoadMaintSA2", Product(), PolicyType(), Gen()).value

    if result is not None:
        return result
    else:
        raise ValueError('lookup failed')


def NetPremRate(basis):
    """Net Premium Rate"""

    gamma2 = LoadMaintSA2
    comf = LifeTable[Sex(), IntRate(basis), TableID(basis)]

    x, n, m = IssueAge(), PolicyTerm(), PremTerm()

    if Product == 'TERM' or Product == 'WL':
        return (comf.Axn(x, n) + gamma2 * comf.AnnDuenx(x, n-m, 1, m)) / comf.AnnDuenx(x, n)

    elif Product == 'ENDW':
        return (comf.Axn(x, n) + gamma2 * comf.AnnDuenx(x, n-m, 1, m)) / comf.AnnDuenx(x, n)

    else:
        raise ValueError('invalid product')


def ReserveNLP_Rate(basis, t):
    """Net level premium reserve rate"""

    gamma2 = LoadMaintSA2()

    lt = LifeTable[Sex(), IntRate(basis), TableID(basis)]

    x, n, m = IssueAge(), PolicyTerm(), PremTerm()

    if t <= m:
        return lt.Axn(x+t, n-t) + gamma2 * lt.AnnDuenx(x+t, n-m, 1, m-t) \
                - NetPremRate(basis) * lt.AnnDuenx(x+t, m-t)
    else:
        return lt.Axn(x+t, n-t) + gamma2 * lt.AnnDuenx(x+t, n-m, 1, m-t)


def ReserveRate():
    """Valuation Reserve Rate per Sum Assured"""
    return None


def SurrCharge(t):
    """Surrender Charge Rate per Sum Assured"""
    m = PremTerm()
    return InitSurrCharge * max((min(m, 10) - t) / min(m, 10), 0)


def TableID(RateBasis):
    """Mortality Table ID"""

    if RateBasis == 'PREM':
        basis = "MortTablePrem"
    elif RateBasis == 'VAL':
        basis = "MortTableVal"
    else:
        raise ValueError('invalid RateBasis')

    result = SpecLookup.match(basis, Product(), PolicyType(), Gen()).value

    if result is not None:
        return result
    else:
        raise ValueError('invalid RateBais')


def UernPremRate():
    """Unearned Premium Rate"""
    return None


Product = lambda: PolicyData[PolicyID, 'Product']

PolicyType = lambda: PolicyData[PolicyID, 'PolicyType']

Gen = lambda: PolicyData[PolicyID, 'Gen']

Channel = lambda: PolicyData[PolicyID, 'Channel']

Sex = lambda: PolicyData[PolicyID, 'Sex']

Duration = lambda: PolicyData[PolicyID, 'Duration']

IssueAge = lambda: PolicyData[PolicyID, 'IssueAge']

PremFreq = lambda: PolicyData[PolicyID, 'PremFreq']

PolicyTerm = lambda: PolicyData[PolicyID, 'PolicyTerm']

PolicyCount = lambda: PolicyData[PolicyID, 'PolicyCount']

SumAssured = lambda: PolicyData[PolicyID, 'SumAssured']

# ---------------------------------------------------------------------------
# References

LifeTable = ("Interface", ("..", "LifeTable"), "auto")

PolicyData = ("Pickle", 2325077945160)

x = ("Interface", (".", "IssueAge"), "auto")

m = ("Interface", (".", "PolicyTerm"), "auto")

PremTerm = ("Interface", (".", "PolicyTerm"), "auto")

n = ("Interface", (".", "PolicyTerm"), "auto")

SpecLookup = ("Interface", ("..", "Input", "SpecLookup"), "auto")


import sys
from lifelib.projects.simplelife.scripts import simplelife

if '' not in sys.path:
    sys.path.insert(0, '')

model = simplelife.build()


def run_simplelife():
    proj = model.Projection
    for i in range(10, 301, 10):
        print(i, proj(i).PV_NetCashflow(0))


if __name__ == '__main__':
    import timeit
    print(timeit.timeit('run_simplelife()', number=1, globals=globals()))



# -*- coding: utf-8 -*-
"""Present Value mix-in modules

This module defines cells that take present values of projected items
in :mod:`projection <simplelife.projection>` module.
"""

def PV_SumInsurIF(t):
    """Present value of insurance in-force"""
    if t > last_t:
        return 0
    else:
        return InsurIF_Beg1(t) + PV_SumInsurIF(t+1) / (1 + DiscRate(t))

def PV_PremIncome(t):
    """Present value of premium income"""
    if t > last_t:
        return 0
    else:
        return PremIncome(t) + PV_PremIncome(t+1) / (1 + DiscRate(t))


def PV_BenefitSurr(t):
    """Present value of surrender benefits"""
    if t > last_t:
        return 0
    else:
        return (-BenefitSurr(t) + PV_BenefitSurr(t+1)) / (1 + DiscRate(t))


def PV_BenefitDeath(t):
    """Present value of death benefits"""
    if t > last_t:
        return 0
    else:
        return (-BenefitDeath(t) + PV_BenefitDeath(t+1)) / (1 + DiscRate(t))


def PV_BenefitMat(t):
    """Present value of matuirty benefits"""
    if t > last_t:
        return 0
    else:
        return (-BenefitMat(t) + PV_BenefitMat(t+1)) / (1 + DiscRate(t))

    
def PV_BenefitTotal(t):
    """Present value of total benefits"""
    if t > last_t:
        return 0
    else:
        return (-BenefitTotal(t) + PV_BenefitTotal(t+1)) / (1 + DiscRate(t))


def PV_ExpsCommTotal(t):
    """Present value of commission expenses"""
    if t > last_t:
        return 0
    else:
        return - ExpsCommTotal(t) + PV_ExpsCommTotal(t+1) / (1 + DiscRate(t))


def PV_ExpsAcq(t):
    """Present value of acquisition expenses"""
    if t > last_t:
        return 0
    else:
        return - ExpsAcq(t) + PV_ExpsAcq(t+1) / (1 + DiscRate(t))
    

def PV_ExpsMaint(t):
    """Present value of maintenance expenses"""
    if t > last_t:
        return 0
    else:
        return - ExpsMaint(t) + PV_ExpsMaint(t+1) / (1 + DiscRate(t))

    
def PV_ExpsTotal(t):
    """Present value of total expenses"""
    if t > last_t:
        return 0
    else:
        return - ExpsTotal(t) + PV_ExpsTotal(t+1) / (1 + DiscRate(t))
    


def PV_NetCashflow(t):
    """Present value of net cashflow"""
    return (PV_PremIncome(t)
            + PV_ExpsTotal(t)
            + PV_BenefitTotal(t))


def PV_NetCashflowForCheck(t):
    """Present value of net cashflow"""
    if t > last_t:
        return 0
    else:
        return (PremIncome(t)
                - ExpsTotal(t)
                - BenefitTotal(t) / (1 + DiscRate(t))
                + PV_NetCashflow(t+1) / (1 + DiscRate(t)))
        
        
def PV_Check(t):
    return PV_NetCashflow(t) - PV_NetCashflowForCheck(t)


def InterestNetCF(t):
    """Interest accreted on pv of net cashflows"""
    if t > last_t:
        return 0
    else:
        return (PV_NetCashflow(t)
                - PremIncome(t)
                + ExpsTotal(t)) * DiscRate(t)


from modelx.serialize.jsonvalues import *

_formula = None

_bases = []

_allow_none = None

_spaces = []

# ---------------------------------------------------------------------------
# References

gen = ("Interface", ("..", "Policy", "Gen"), "auto")

polt = ("Interface", ("..", "Policy", "PolicyType"), "auto")

prod = ("Interface", ("..", "Policy", "Product"), "auto")

sex = ("Interface", ("..", "Policy", "Sex"), "auto")


"""Space for cashflow projection.

This Space is for projecting cashflows.
Most Cells in this Spaces return calculation results for all model points as
pandas Series objects indexed by Policy ID.

This Space has child Spaces,
:mod:`~fastlife.model.Projection.Policy` and :mod:`~fastlife.model.Projection.Assumptions`.
The :mod:`~fastlife.model.Projection.Policy` Space contains Cells representing policy attributes, such as
product type, issue age, sum assured, etc.
It also contains Cells for calculating policy values such as premium rates and
cash surrender value rates.

This Space has a base Space, :mod:`~fastlife.model.PV`, in which
Cells for taking present values of cashflows are defined.

.. rubric:: Composition Structure

.. blockdiag::

   blockdiag {
     default_node_color="#D5E8D4";
     default_linecolor="#628E47";
     node_width=150;
     Projection <- Assumptions [hstyle=composition];
     Projection <- Policy [hstyle=composition];
   }

.. rubric:: Inheritance Structure

.. blockdiag::

   blockdiag {
     default_node_color="#D5E8D4";
     default_linecolor="#628E47";
     PV[style=dotted]
     PV<- Projection [hstyle=generalization];
   }

.. rubric:: References

The following attributes defined in this Space.

Attributes:
    pol: Alias for :mod:`~fastlife.model.Projection.Policy` child Space
    asmp: Alias for :mod:`~fastlife.model.Projection.Assumptions` child Space
    scen: Alias for :mod:`~fastlife.model.Economic` Space
    ScenID: Scenario ID (``1`` by default)


"""

from modelx.serialize.jsonvalues import *

_formula = lambda PolicyID, ScenID=1: None

_bases = [
    ".PV"
]

_allow_none = None

_spaces = [
    "Policy",
    "Assumptions"
]

# ---------------------------------------------------------------------------
# Cells

def DiscRate(t):
    """Rates for discount cashflows

    Refers to :func:`Economic[ScenID].DiscRate<simplelife.model.Economic.DiscRate>`
    """
    return scen[ScenID].DiscRate(t)


def InflFactor(t):
    """Inflation factors to adjust expense cashflows

    Refers to :func:`Economic[ScenID].InflFactor<simplelife.model.Economic.InflFactor>`
    """
    return scen[ScenID].InflFactor(t)


def InvstRetRate(t):
    """Rate of investment return

    Refers to :func:`Economic[ScenID].InvstRetRate<simplelife.model.Economic.InvstRetRate>`
    """
    return scen[ScenID].InvstRetRate(t)


def BaseMortRate(t):
    """Base mortality rate"""

    exist = (t <= last_t())

    keys = pd.concat([asmp.MortTableID(), pol.Sex(), AttAge(t), exist],
                      axis=1, keys=["ID", "Sex", "Age", "exist"])

    def find_rate(key):
        if key["exist"]:
            return asmp.MortalityTables()[key["ID"], key["Sex"]][key["Age"]]
        else:
            return 0

    result = keys.apply(find_rate, axis=1)
    result.name = "BaseMortRate"

    return result


def last_t():

    result = np.minimum(asmp.LastAge() - pol.IssueAge(), pol.PolicyTerm())
    result.name = "last_t"

    return result


def AccumCF(t):
    """Accumulated cashflows"""
    if t == 0:
        return 0
    else:
        return (AccumCF(t-1)
                + IntAccumCF(t-1)
                + NetInsurCF(t-1))


def AttAge(t):
    """Attained age at time ``t``"""
    return pol.IssueAge + t


def BenefitAccDth(t):
    """Accidental death benefits"""
    return SizeBenefitAccDth(t) * PolsAccDeath(t)


def BenefitAccHosp(t):
    """Accidental hospitalization benefits"""
    return SizeBenefitAccHosp(t) * PolsAccHosp(t)


def BenefitAnn(t):
    """Annuity benefits"""
    return SizeBenefitAnn(t) * PolsAnnuity(t)


def BenefitDeath(t):
    """Death benefits"""
    return SizeBenefitDeath(t) * PolsDeath(t)


def BenefitLiving(t):
    """Living benefits"""
    return SizeBenefitLiving(t) * PolsLiving(t)


def BenefitMat(t):
    """Matuirty benefits"""
    return SizeBenefitMat(t) * PolsMaturity(t)


def BenefitOther(t):
    """Other benefits"""
    return SizeBenefitOther(t) * PolsOther(t)


def BenefitSickHosp(t):
    """Sickness hospitalization benefits"""
    return SizeBenefitSickHosp(t) * PolsSickHosp(t)


def BenefitSurg(t):
    """Surgery benefits"""
    return SizeBenefitSurg(t) * PolsSurg(t)


def BenefitSurr(t):
    """Surrender benefits"""
    return SizeBenefitSurr(t) * PolsSurr(t)


def BenefitTotal(t):
    """Benefit Total"""
    return (BenefitMat(t)
            + BenefitDeath(t)
            + BenefitAccDth(t)
            + BenefitSurr(t)
            + BenefitAnn(t)
            + BenefitAccHosp(t)
            + BenefitSickHosp(t)
            + BenefitSurg(t)
            + BenefitLiving(t)
            + BenefitOther(t))


def ChangeRsrv(t):
    """Change in reserve"""
    return ReserveTotal_End(t+1) - ReserveTotal_End(t)


def ExpsAcq(t):
    """Acquisition expenses"""
    return SizeExpsAcq(t) * (PolsNewBiz(t) + PolsRenewal(t))


def ExpsAcqTotal(t):
    """Commissions and acquisition expenses"""
    return ExpsCommTotal(t) + ExpsAcq(t)


def ExpsCommInit(t):
    """Initial commissions"""
    return SizeExpsCommInit(t) * PolsIF_Beg1(t)


def ExpsCommRen(t):
    """Renewal commissions"""
    return SizeExpsCommRen(t) * PolsIF_Beg1(t)


def ExpsCommTotal(t):
    """Commissions Total"""
    return ExpsCommInit(t) + ExpsCommRen(t)


def ExpsMaint(t):
    """Maintenance expenses"""
    return SizeExpsMaint(t) * PolsIF_Beg1(t)


def ExpsMaintTotal(t):
    """Total maintenance expenses including other expenses"""
    return ExpsMaint(t) + ExpsOther(t)


def ExpsOther(t):
    """Other expenses"""
    return 0


def ExpsTotal(t):
    """Total expenses"""
    return (ExpsCommInit(t)
            + ExpsCommRen(t)
            + ExpsAcq(t)
            + ExpsMaint(t)
            + ExpsOther(t))


def IncomeTotal(t):
    """Income Total"""
    return PremIncome(t) + InvstIncome(t)


def InsurIF_Beg1(t):
    """Insurance in-force: Beginning of period 1"""
    return PolsIF_Beg1(t) * SizeSumAssured(t)


def InsurIF_End(t):
    """Insurance in-force: End of period"""
    return PolsIF_End(t) * SizeSumAssured(t)


def IntAccumCF(t):
    """Intrest on accumulated cashflows"""
    return (AccumCF(t)
            + PremIncome(t)
            - ExpsTotal(t)) * DiscRate(t)


def InvstIncome(t):
    """Investment income"""
    return SizeInvstIncome(t) * PolsIF_Beg1(t)


def NetInsurCF(t):
    """Net liability cashflow"""
    return (PremIncome(t)
            - BenefitTotal(t)
            - ExpsTotal(t))


def PolsAccDeath(t):
    """Number of policies: Accidental death"""
    return 0


def PolsAccHosp(t):
    """Number of policies: Accidental Hospitalization"""
    return 0


def PolsAnnuity(t):
    """Number of policies: Annuity"""
    return 0


def PolsDeath(t):
    """Number of policies: Death"""
    return PolsIF_Beg1(t) * BaseMortRate(t) * asmp.MortFactor(t)


def PolsIF_AftMat(t):
    """Number of policies: Maturity"""
    return PolsIF_End(t) - PolsMaturity(t)


def PolsIF_Beg(t):
    """Number of policies: Beginning of period"""
    return PolsIF_AftMat(t)


def PolsIF_Beg1(t):
    """Number of policies: Beginning of period 1"""
    return PolsIF_Beg(t) + PolsRenewal(t) + PolsNewBiz(t)


def PolsIF_End(t):
    """Number of policies: End of period"""
    if t == 0:
        return 0 # pol.PolicyCount()
    else:
        return PolsIF_Beg1(t-1) - PolsDeath(t-1) - PolsSurr(t-1)


def PolsLiving(t):
    """Number of policies: Living benefits"""
    return 0


def PolsMaturity(t):
    """Number of policies: Maturity"""
    return (pol.PolicyTerm == t) * PolsIF_End(t)


def PolsNewBiz(t):
    """Number of policies: New business"""
    return pol.PolicyCount() if t == 0 else 0


def PolsOther(t):
    """Number of policies: Other benefits"""
    return 0


def PolsRenewal(t):
    """Number of policies: Renewal policies"""
    return 0


def PolsSickHosp(t):
    """Number of policies: Sickness Hospitalization"""
    return 0


def PolsSurg(t):
    """Number of policies: Surgery"""
    return 0


def PolsSurr(t):
    """Number of policies: Surrender"""
    return PolsIF_Beg1(t) * asmp.SurrRate(t)


def PremIncome(t):
    """Premium income"""
    return SizePremium(t) * PolsIF_Beg1(t)


def ProfitBefTax(t):
    """Profit before Tax"""

    return (PremIncome(t)
            + InvstIncome(t)
            - BenefitTotal(t)
            - ExpsTotal(t)
            - ChangeRsrv(t))


def ReserveHospRsrvEnd(t):
    """Hospitalization reserve: End of period"""
    return 0


def ReservePremRsrvEnd(t):
    """Premium reserve: End of period"""
    return SizeReservePremRsrvEnd(t) * PolsIF_End(t)


def ReserveTotal_End(t):
    """Total reserve: End of period"""
    return (ReservePremRsrvEnd(t)
            + ReserveUernPremEnd(t)
            + ReserveHospRsrvEnd(t))


def ReserveUernPremEnd(t):
    """Unearned Premium: End of period"""
    return 0


def SizeAnnPrem(t):
    """Annualized premium per policy at time ``t``"""
    return SizeSumAssured(t) * pol.AnnPremRate()


def SizeBenefitAccDth(t):
    """Accidental death benefit per policy"""
    return 0


def SizeBenefitAccHosp(t):
    """Accidental hospitalization benefit per policy"""
    return 0


def SizeBenefitAnn(t):
    """Annuity benefit per policy"""
    return 0


def SizeBenefitDeath(t):
    """Death benefit per policy"""
    return SizeSumAssured(t)


def SizeBenefitLiving(t):
    """Living benefit per policy"""
    return 0


def SizeBenefitMat(t):
    """Maturity benefit per policy"""
    return 0


def SizeBenefitOther(t):
    """Other benefit per policy"""
    return 0


def SizeBenefitSickHosp(t):
    """Sickness hospitalization benefit per policy"""
    return 0


def SizeBenefitSurg(t):
    """Surgery benefit per policy"""
    return 0


def SizeBenefitSurr(t):
    """Surrender benefit per policy"""
    return SizeSumAssured(t) * (pol.CashValueRate(t)
                                + pol.CashValueRate(t+1)) / 2


def SizeExpsAcq(t):
    """Acquisition expense per policy at time t"""
    if t == 0:
        return (SizeAnnPrem(t) * asmp.ExpsAcqAnnPrem()
                + (SizeSumAssured(t) * asmp.ExpsAcqSA() + asmp.ExpsAcqPol())
                * InflFactor(t) / InflFactor(0))
    else:
        return 0


def SizeExpsCommInit(t):
    """Initial commission per policy at time t"""
    if t == 0:
        return SizePremium(t) * asmp.CommInitPrem() * (1 + asmp.CnsmpTax)
    else:
        return 0


def SizeExpsCommRen(t):
    """Renewal commission per policy at time t"""
    if t == 0:
        return 0
    elif (t < asmp.CommRenTerm()).any():
        return (t < asmp.CommRenTerm()) * SizePremium(t) * asmp.CommRenPrem() * (1 + asmp.CnsmpTax())
    else:
        return 0


def SizeExpsMaint(t):
    """Maintenance expense per policy at time t"""
    return (SizeAnnPrem(t) * asmp.ExpsMaintAnnPrem()
            + (SizeSumAssured(t) * asmp.ExpsMaintSA() + asmp.ExpsMaintPol())
            * InflFactor(t))


def SizeExpsOther(t):
    """Other expenses per policy at time t"""
    return 0


def SizeInvstIncome(t):
    """Investment Income per policy from t to t+1"""
    return (SizeReserveTotalAftMat(t) + SizePremium(t)) * InvstRetRate(t)


def SizePremium(t):
    """Premium income per policy from t to t+1"""
    return SizeSumAssured(t) * pol.GrossPremRate() * pol.PremFreq()


def SizeReservePremRsrvAftMat(t):
    """Premium reserve per policy: After maturity"""
    return SizeSumAssured(t) * pol.ReserveNLP_Rate('VAL', t)


def SizeReservePremRsrvEnd(t):
    """Premium reserve per policy: End of period"""
    return SizeSumAssured(t) * pol.ReserveNLP_Rate('VAL', t)


def SizeReserveTotalAftMat(t):
    """Total reserve per policy: After maturity"""
    return (SizeReservePremRsrvAftMat(t)
           + SizeReserveUernPremAftMat(t))


def SizeReserveUernPremAftMat(t):
    """Unearned premium: After maturity"""
    return 0 # SizeSumAssured(t) * polset.UnernPremRate(polset, tt, True)


def SizeReserveUernPremEnd(t):
    """Unearned reserve per policy: End of period"""
    return 0 # SizeSumAssured(t) * pol.UnernPremRate(polset, tt)


def SizeSumAssured(t):
    """Sum assured per policy at time ``t``"""
    return  pol.SumAssured()


def InterestNetCF(t):
    """Interest accreted on pv of net cashflows"""
    if t > last_t():
        return 0
    else:
        return (PV_NetCashflow(t)
                - PremIncome(t)
                + ExpsTotal(t)) * DiscRate(t)


def PV_BenefitDeath(t):
    """Present value of death benefits"""
    if t > last_t():
        return 0
    else:
        return (-BenefitDeath(t) + PV_BenefitDeath(t+1)) / (1 + DiscRate(t))


def PV_BenefitMat(t):
    """Present value of matuirty benefits"""
    if t > last_t():
        return 0
    else:
        return (-BenefitMat(t) + PV_BenefitMat(t+1)) / (1 + DiscRate(t))


def PV_BenefitSurr(t):
    """Present value of surrender benefits"""
    if t > last_t():
        return 0
    else:
        return (-BenefitSurr(t) + PV_BenefitSurr(t+1)) / (1 + DiscRate(t))


def PV_ExpsAcq(t):
    """Present value of acquisition expenses"""
    if t > last_t():
        return 0
    else:
        return - ExpsAcq(t) + PV_ExpsAcq(t+1) / (1 + DiscRate(t))


def PV_ExpsCommTotal(t):
    """Present value of commission expenses"""
    if t > last_t():
        return 0
    else:
        return - ExpsCommTotal(t) + PV_ExpsCommTotal(t+1) / (1 + DiscRate(t))


def PV_ExpsMaint(t):
    """Present value of maintenance expenses"""
    if t > last_t():
        return 0
    else:
        return - ExpsMaint(t) + PV_ExpsMaint(t+1) / (1 + DiscRate(t))


def PV_NetCashflowForCheck(t):
    """Present value of net cashflow"""
    if t > last_t():
        return 0
    else:
        return (PremIncome(t)
                - ExpsTotal(t)
                - BenefitTotal(t) / (1 + DiscRate(t))
                + PV_NetCashflow(t+1) / (1 + DiscRate(t)))


def PV_SumInsurIF(t):
    """Present value of insurance in-force"""
    if t > last_t():
        return 0
    else:
        return InsurIF_Beg1(t) + PV_SumInsurIF(t+1) / (1 + DiscRate(t))


# ---------------------------------------------------------------------------
# References

pol = ("Interface", (".", "Policy"), "auto")

asmp = ("Interface", (".", "Assumptions"), "auto")

scen = ("Interface", ("..", "Economic"), "auto")

ScenID = 1


"""The main module to build a simplelife model.

This module contains only one function :py:func:`build`,
which creates a model from source modules and return it.

If this module is run as a script, the :py:func:`build` function is called
and the created model is available as ``model`` global variable.
"""
import os
import modelx as mx

# %% Code block for overriding lapse logic.

def SurrRateMult_outer(t):
    """Surrender rate multiple for the outer projection (Default: 1)"""
    if t == 0:
        return 1
    else:
        return SurrRateMult(t-1)


def SurrRateMult_inner(t):
    """Surrender rate multiple for the inner projection (Default: 1)"""
    if t == 0:
        return outer.SurrRateMult(t)

    elif t == t0:
        return _space.parent(t-1).SurrRateMult(t-1)

    else:
        return SurrRateMult(t-1)


def PolsSurr(t):
    """Number of policies: Surrender"""    
    return PolsIF_Beg1(t) * asmp.SurrRate(t) * SurrRateMult(t)


def PolsIF_End_inner(t):
    """Number of policies: End of period"""
    if t == t0:
        return outer.PolsIF_End(t)
    else:
        return PolsIF_Beg1(t-1) - PolsDeath(t-1) - PolsSurr(t-1)

# %% Code block for overriding discounting logic.

def IntAccumCF_outer(t):
    """Intrest on accumulated cashflows"""
    return (AccumCF(t)
            + PremIncome(t)
            - ExpsTotal(t)) * DiscRate(t, 0)


def IntAccumCF_inner(t):
    """Intrest on accumulated cashflows"""
    return (AccumCF(t)
            + PremIncome(t)
            - ExpsTotal(t)) * DiscRate(t)


def DiscRateAdj(t):
    """Adjustment to the outer discount rates"""
    if t == 0:
        return 0
    else:
        return DiscRateAdj(t-1)

    
def DiscRate_outer(t, dur):
    """Discount rates for the outer projection"""
    return scen.DiscRate(dur) + DiscRateAdj(t)


def DiscRate_inner(t):
    """Discount rates for the inner projection"""
    return outer.DiscRate(t0, t)
    
    
# %% Code block for build function

def build(load_saved=False):
    """Build a model and return it.

    Read input data from `input.xlsx`, create `Input` space and its
    subspace and cells and populate them with the data.

    Args:
        load_saved: If ``True``, input data is read from `ifrs17sim.mx` file
            instead of `input.xlsx`, which is saved when
            :py:func:`build_input <simplelife.build_input.build_input>`
            is executed last time. Defaults to ``False``
    """

    # Make sure the current directory is this folder
    os.chdir(os.path.abspath(os.path.dirname(__file__)))

    # ------------------------------------------------------------------------
    # Build Input space

    from build_input import build_input

    if load_saved:
        model = mx.open_model('ifrs17sim.mx')
        input = model.Input
    else:
        model = mx.new_model(name='ifrs17sim')
        input = build_input(model, 'input.xlsx')
        model.save('ifrs17sim.mx')

    # ------------------------------------------------------------------------
    # Build CommFunc space

    lifetable_refs = {'Input': input}

    def lifetable_params(Sex, IntRate, TableID):
        refs={'MortalityTable': Input.MortalityTables(TableID).MortalityTable}
        return {'refs': refs}

    lifetable = model.import_module(
        module='lifetable',
        name='LifeTable',
        formula=lifetable_params,
        refs=lifetable_refs)

    # ------------------------------------------------------------------------
    # Build Policy space

    from policy import policy_attrs

    policy_refs = {'PolicyData': input.PolicyData,
                   'ProductSpec': input.ProductSpec,
                   'LifeTable': lifetable,
                   'PolicyAttrs': policy_attrs}

    def policy_params(PolicyID):
        refs = {attr: PolicyData[PolicyID].cells[attr]() for attr in PolicyAttrs}
        alias = {'PremTerm': refs['PolicyTerm'],
                 'x': refs['IssueAge'],
                 'm': refs['PolicyTerm'],
                 'n': refs['PolicyTerm']}

        refs.update(alias)
        return {'refs': refs}

    policy = model.import_module(
        module='policy',
        name='Policy',
        formula=policy_params,
        refs=policy_refs)

    # ------------------------------------------------------------------------
    # Build Assumption space

    asmp_refs = {'Policy': policy,
                 'ProductSpec': input.ProductSpec,
                 'MortalityTables': input.MortalityTables,
                 'asmp': input.Assumption,
                 'asmp_tbl': input.AssumptionTables}

    def asmp_params(PolicyID):
        refs = {'pol': Policy[PolicyID]}
        alias = {'prod': refs['pol'].Product,
                 'polt': refs['pol'].PolicyType,
                 'gen': refs['pol'].Gen}
        refs.update(alias)
        return {'refs': refs}

    asmp = model.import_module(
        module='assumption',
        name='Assumption',
        formula=asmp_params,
        refs=asmp_refs)

    asmp.allow_none = True

    # ------------------------------------------------------------------------
    # Build Assumption space

    def econ_params(ScenID):
        refs = {'Scenario': Input.Scenarios[ScenID]}
        return {'refs': refs}

    economic = model.import_module(
        module='economic',
        name='Economic',
        formula=econ_params,
        refs={'asmp': asmp,
              'Input': input})

    # ------------------------------------------------------------------------
    # Build Projection space
    
    # Model tree structure
    # 
    # lifelib --+
    #           +--BaseProj
    #           +--OuterProj[PolicyID] <--- BaseProj
    #                    +--InnerProj[t] <-- BaseProj

    proj_refs = {'Pol': policy,
                 'Asmp': asmp,
                 'Scen': economic}

    def proj_params(PolicyID, ScenID=1):
        refs = {'pol': Pol[PolicyID],
                'asmp': Asmp[PolicyID],
                'scen': Scen[ScenID]}

        return {'refs': refs}

    baseproj = model.import_module(
        module='projection',
        name='BaseProj')

    ifrs = model.import_module(
        module='ifrs',
        name='IFRS')

    outerproj = model.new_space(
        bases=[ifrs, baseproj],
        name='OuterProj',
        formula=proj_params,
        refs=proj_refs)

    def innerproj_params(t0):
        refs = {'pol': _space.parent.pol,
                'asmp': _space.parent.asmp,
                'scen': _space.parent.scen,
                'outer': _space.parent}
        
        return {'refs': refs}

    innerproj = outerproj.new_space(
        bases=baseproj,
        name='InnerProj',
        formula=innerproj_params)

    pvs = innerproj.import_module(
        module='present_value',
        name='PresentValue')
    
    def pvs_params(t_rate):
        refs = {'last_t': _space.parent.last_t,
                'InsurIF_Beg1': _space.parent.InsurIF_Beg1,
                'InsurIF_End': _space.parent.InsurIF_End,
                'PremIncome': _space.parent.PremIncome,
                'BenefitSurr': _space.parent.BenefitSurr,
                'BenefitDeath': _space.parent.BenefitDeath,
                'BenefitTotal': _space.parent.BenefitTotal,
                'ExpsCommTotal': _space.parent.ExpsCommTotal,
                'ExpsAcq': _space.parent.ExpsAcq,
                'ExpsMaint': _space.parent.ExpsMaint,
                'ExpsTotal': _space.parent.ExpsTotal,
                'DiscRate': _space.parent.parent[t_rate].DiscRate}
        
        return {'refs': refs}
        
    pvs.set_formula(pvs_params)
    
    # Add or override functions.
    baseproj.new_cells(name='SurrRateMult', formula=SurrRateMult_outer)
    baseproj.PolsSurr.set_formula(PolsSurr)
    outerproj.IntAccumCF.set_formula(IntAccumCF_outer)
    outerproj.new_cells(name='DiscRate', formula=DiscRate_outer)
    outerproj.new_cells(formula=DiscRateAdj)
    innerproj.new_cells(name='DiscRate', formula=DiscRate_inner)
    innerproj.SurrRateMult.set_formula(SurrRateMult_inner)
    innerproj.IntAccumCF.set_formula(IntAccumCF_inner)
    innerproj.PolsIF_End.set_formula(PolsIF_End_inner)
    
    return model


if __name__ == '__main__':
    model = build()



"""Source module to create ``LifeTable`` space from.

This is a source module to create ``LifeTable`` space and its
sub spaces from.

This module is passed to ``import_module`` method to create
a space that contains cells that defines life tables and commutation functions,
for a selected combination of ``Sex``, ``IntRate`` and ``MortalityTable``.

``MortalityTable`` and ``Sex`` are used in :py:func:`qx` below to identify
the mortality rates to be applied.

Example
    Sample script::

        from modelx import *
        space = new_model().import_module(module=lifetable)
        space.Sex = 'M'
        space.IntRate = 0.03
        space.MortalityTable = lambda sex, x: 0.001 if x < 110 else 1

References:
    * `International actuarial notation by F.S.Perryman <https://www.casact.org/pubs/proceed/proceed49/49123.pdf>`_

.. rubric:: Project Templates

This module is included in the following project templates.

* :mod:`simplelife`
* :mod:`nestedlife`

.. rubric:: References in Sub

Attributes:
    Sex: 'M' or 'F' to indicate male or female column in the mortality table.
    IntRate: The constant interest rate for discounting.
    MortalityTable: The ultimate mortality table by sex and age.

"""


def disc():
    """The discount factor :math:`v = 1/(1 + i)`."""
    return 1 / (1 + IntRate)


def lx(x):
    """The number of persons remaining at age ``x``. """
    if x == 0:
        return 100000
    else:
        return lx(x-1) - dx(x-1)

def dx(x):
    """The number of persons who die between ages ``x`` and ``x+1``"""
    return lx(x) * qx(x)


def qx(x):
    """Probability that a person at age ``x`` will die in one year."""
    return MortalityTable(Sex, x)


def Dx(x):
    """The commutation column :math:`D_{x} = l_{x}v^{x}`.
    """
    return lx(x) * disc ** x


def Cx(x):
    """The commutation column :math:`\\overline{C_x}`.
    """

    return dx(x) * disc**(x+1/2)


def Nx(x):
    """The commutation column :math:`N_x`."""
    if x >= 110:    # TODO: Get the last age from the table
        return Dx(x)
    else:
        return Nx(x+1) + Dx(x)


def Mx(x):
    """The commutation column :math:`M_x`."""

    if x >= 110:
        return Dx(x)
    else:
        return Mx(x+1) + Cx(x)


def Ax(x, f=0):
    """The present value of a lifetime assurance on a person at age ``x``
    payable immediately upon death, optionally with an waiting period of ``f`` years.

    .. math::

        \\require{enclose}{}_{f|}\\overline{A}_{x}
    """
    if Dx(x) == 0:
        return 0
    else:
        return Mx(x+f) / Dx(x)


def Axn(x, n, f=0):
    """The present value of an assurance on a person at age ``x`` payable
    immediately upon death, optionally with an waiting period of ``f`` years.

    .. math::

        \\require{enclose}{}_{f|}\\overline{A}^{1}_{x:\\enclose{actuarial}{n}}

    """
    if Dx(x) == 0:
        return 0
    else:
        return (Mx(x+f) - Mx(x+f+n)) / Dx(x)


def Exn(x, n):
    """ The value of an endowment on a person at age ``x``
    payable after n years

    .. math::

        {}_{n}E_x

    """
    if Dx(x) == 0:
        return 0
    else:
        return Dx(x+n) / Dx(x)


def AnnDuenx(x, n, k=1, f=0):
    """ The present value of an annuity-due.

    .. math::

        \\require{enclose}{}_{f|}\\ddot{a}_{x:\\enclose{actuarial}{n}}^{(k)}


    Args:
        x(int): age
        n(int): length of annuity payments in years
        k(int, optional): number of split payments in a year
        f(int, optional): waiting period in years

    """
    if Dx(x) == 0:
        return 0
    
    result = (Nx(x+f) - Nx(x+f+n)) / Dx(x)

    if k > 1:
        return result - (k-1) / (2*k) * (1 - Dx(x+f+n) / Dx(x))
    else:
        return result


def AnnDuex(x, k, f=0):
    """The present value of a lifetime annuity due.

    Args:
        x(int): age
        k(int, optional): number of split payments in a year
        f(int, optional): waiting period in years
    """
    if Dx(x) == 0:
        return 0
    
    result = (Nx(x+f)) / Dx(x)

    if k > 1:
        return result - (k-1) / (2*k)
    else:
        return result



import unittest
import smithwilson as sw
import numpy as np

class TestSmithWilson(unittest.TestCase):

    def test_ufr_discount_factor(self):
        """Test creation of UFR discount factor vector"""

        # Input
        ufr = 0.029
        t = np.array([0.25, 1.0, 5.0, 49.5, 125.0])

        # Expected Output
        expected = np.array([0.992878614, 0.971817298, 0.866808430, 0.242906395, 0.028059385])

        # Actual Output
        actual = sw.ufr_discount_factor(ufr=ufr, t=t)

        # Assert
        self.assertEqual(type(actual), type(expected), "Returned types not matching")
        self.assertTupleEqual(actual.shape, expected.shape, "Shapes not matching")
        np.testing.assert_almost_equal(actual, expected, decimal=8, err_msg="UFR discount factors not matching")


    def test_calculate_prices(self):
        """Test calculation of zero-coupon bond price vector"""

        # Input
        r = np.array([0.02, 0.025, -0.033, 0.01, 0.0008])
        t = np.array([0.25, 1.0, 5.0, 49.5, 125.0])

        # Expected Output
        expected = np.array([0.995061577, 0.975609756, 1.182681027, 0.611071456, 0.904873593])

        # Actual Output
        actual = sw.calculate_prices(rates=r, t=t)

        # Assert
        self.assertEqual(type(actual), type(expected), "Returned types not matching")
        self.assertTupleEqual(actual.shape, expected.shape, "Shapes not matching")
        np.testing.assert_almost_equal(actual, expected, decimal=8, err_msg="Prices not matching")


    def test_wilson_function_symmetric(self):
        """Test creation of a symmetric Wilson-function matrix (t1 = t2)"""

        # Input
        t = np.array([0.25, 1.0, 5.0, 49.5, 125.0]).reshape((-1, 1))
        ufr = 0.029
        alpha = 0.2

        # Expected Output
        expected = np.array([[0.00238438, 0.00872884, 0.02719467, 0.01205822, 0.00139298],
                             [0.00872884, 0.03320614, 0.10608305, 0.04720974, 0.00545372],
                             [0.02719467, 0.10608305, 0.42652097, 0.2105409 , 0.02432211],
                             [0.01205822, 0.04720974, 0.2105409 , 0.55463306, 0.06747646],
                             [0.00139298, 0.00545372, 0.02432211, 0.06747646, 0.01928956]])

        # Actual Output
        actual = sw.wilson_function(t1=t, t2=t, ufr=ufr, alpha=alpha)

        # Assert
        self.assertEqual(type(actual), type(expected), "Returned types not matching")
        self.assertTupleEqual(actual.shape, expected.shape, "Shapes not matching")
        np.testing.assert_almost_equal(actual, expected, decimal=8, err_msg="Wilson functions not matching")


    def test_wilson_function_asymmetric_t1_lt_t2(self):
        """Test creation of a symmetric Wilson-function matrix (t1 != t2) with length of t1 > length of t2"""

        # Input
        t_obs = np.array([0.25, 1.0, 5.0, 49.5, 125.0]).reshape((-1, 1))
        t_target = np.array([0.25, 0.5, 1.0, 2.0, 2.5, 3.5, 5.0, 10.0, 20.0, 49.5, 125.0]).reshape((-1, 1))
        ufr = 0.029
        alpha = 0.2

        # Expected Output
        expected = np.array([[0.00238438, 0.00872884, 0.02719467, 0.01205822, 0.00139298],
                             [0.00463874, 0.01723526, 0.0539627 , 0.0239447 , 0.00276612],
                             [0.00872884, 0.03320614, 0.10608305, 0.04720974, 0.00545372],
                             [0.015444  , 0.05969492, 0.20375322, 0.0917584 , 0.01060004],
                             [0.01817438, 0.07046799, 0.24880429, 0.11307011, 0.013062  ],
                             [0.02260267, 0.08794588, 0.33012767, 0.15383656, 0.01777143],
                             [0.02719467, 0.10608305, 0.42652097, 0.2105409 , 0.02432211],
                             [0.03225016, 0.12614043, 0.54769846, 0.36498556, 0.04216522],
                             [0.02751232, 0.10770227, 0.47881259, 0.54833094, 0.06336226],
                             [0.01205822, 0.04720974, 0.2105409 , 0.55463306, 0.06747646],
                             [0.00139298, 0.00545372, 0.02432211, 0.06747646, 0.01928956]])

        # Actual Output
        actual = sw.wilson_function(t1=t_target, t2=t_obs, ufr=ufr, alpha=alpha)

        # Assert
        self.assertEqual(type(actual), type(expected), "Returned types not matching")
        self.assertTupleEqual(actual.shape, expected.shape, "Shapes not matching")
        np.testing.assert_almost_equal(actual, expected, decimal=8, err_msg="Wilson functions not matching")


    def test_wilson_function_asymmetric_t2_lt_t1(self):
        """Test creation of a symmetric Wilson-function matrix (t1 != t2) with length of t2 > length of t1"""

        # Input
        t_target = np.array([0.50, 1.5, 7.0, 22.5]).reshape((-1, 1))
        t_obs = np.array([0.25, 1.0, 2.0, 2.5, 5.0, 10.0, 20.0]).reshape((-1, 1))
        ufr = 0.032
        alpha = 0.15

        # Expected Output
        expected = np.array([[0.00263839, 0.00990704, 0.01791847, 0.02129457, 0.03324991, 0.04184617, 0.03736174],
                             [0.00714378, 0.02751832, 0.05096578, 0.06087744, 0.09600535, 0.12138299, 0.1085669 ],
                             [0.01939785, 0.07563626, 0.14568738, 0.17843321, 0.31674624, 0.45088288, 0.42190812],
                             [0.01768861, 0.06909389, 0.13384921, 0.16464728, 0.3035725 , 0.51271549, 0.69668792]])

        # Actual Output
        actual = sw.wilson_function(t1=t_target, t2=t_obs, ufr=ufr, alpha=alpha)

        # Assert
        self.assertEqual(type(actual), type(expected), "Returned types not matching")
        self.assertTupleEqual(actual.shape, expected.shape, "Shapes not matching")
        np.testing.assert_almost_equal(actual, expected, decimal=8, err_msg="Wilson functions not matching")


    def test_fit_parameters(self):
        """Test estimation of Smith-Wilson parameter vector ζ"""

        # Input
        r = np.array([0.02, 0.025, -0.033, 0.01, 0.0008]).reshape((-1, 1))
        t = np.array([0.25, 1.0, 5.0, 49.5, 125.0]).reshape((-1, 1))
        ufr = 0.029
        alpha = 0.2

        # Expected Output
        expected = np.array([-42.78076209, 23.4627511, -3.96498616, 8.92604195, -75.22418515]).reshape((-1, 1))

        # Actual Output
        actual = sw.fit_parameters(rates=r, t=t, ufr=ufr, alpha=alpha)

        # Assert
        self.assertEqual(type(actual), type(expected), "Returned types not matching")
        self.assertTupleEqual(actual.shape, expected.shape, "Shapes not matching")
        np.testing.assert_almost_equal(actual, expected, decimal=8, err_msg="Parameter not matching")


    def test_fit_smithwilson_rates_actual(self):
        """Test estimation of yield curve fitted with the Smith-Wilson algorithm.
           This example uses an actual example from EIOPA. Deviations must be less than 1bps (0.01%).
           Source: https://wayback.archive-it.org/org-1495/20191229100044/https:/eiopa.europa.eu/Publications/Standards/EIOPA_RFR_20190531.zip
                   EIOPA_RFR_20190531_Term_Structures.xlsx; Tab: RFR_spot_no_VA; Switzerland
        """

        # Input
        r = np.array([-0.00803, -0.00814, -0.00778, -0.00725, -0.00652,
                      -0.00565, -0.0048, -0.00391, -0.00313, -0.00214,
                      -0.0014, -0.00067, -0.00008, 0.00051, 0.00108,
                      0.00157, 0.00197, 0.00228, 0.0025, 0.00264,
                      0.00271, 0.00274, 0.0028, 0.00291, 0.00309]).reshape((-1, 1))
        t = np.array([float(y + 1) for y in range(len(r))]).reshape((-1, 1)) # 1.0, 2.0, ..., 25.0
        ufr = 0.029
        alpha = 0.128562

        t_target = np.array([float(y + 1) for y in range(65)]).reshape((-1, 1))

        # Expected Output
        expected = np.array([-0.00803, -0.00814, -0.00778, -0.00725, -0.00652,
                             -0.00565, -0.0048, -0.00391, -0.00313, -0.00214,
                             -0.0014, -0.00067, -0.00008, 0.00051, 0.00108,
                             0.00157, 0.00197, 0.00228, 0.0025, 0.00264,
                             0.00271, 0.00274, 0.0028, 0.00291, 0.00309,
                             0.00337, 0.00372, 0.00412, 0.00455, 0.00501,
                             0.00548, 0.00596, 0.00644, 0.00692, 0.00739,
                             0.00786, 0.00831, 0.00876, 0.00919, 0.00961,
                             0.01002, 0.01042, 0.01081, 0.01118, 0.01154,
                             0.01189, 0.01223, 0.01255, 0.01287, 0.01318,
                             0.01347, 0.01376, 0.01403, 0.0143, 0.01456,
                             0.01481, 0.01505, 0.01528, 0.01551, 0.01573,
                             0.01594, 0.01615, 0.01635, 0.01655, 0.01673]).reshape((-1, 1))

        # Actual Output
        actual = sw.fit_smithwilson_rates(rates_obs=r, t_obs=t, t_target=t_target, ufr=ufr, alpha=alpha)

        # Assert - Precision of 4 decimal points equals deviatino of less than 1bps
        self.assertEqual(type(actual), type(expected), "Returned types not matching")
        self.assertTupleEqual(actual.shape, expected.shape, "Shapes not matching")
        np.testing.assert_almost_equal(actual, expected, decimal=4, err_msg="Fitted rates not matching")


    def test_fit_smithwilson_rates_random(self):
        """Test estimation of yield curve fitted with the Smith-Wilson algorithm
           This test uses random data points.
        """

        # Input
        r = np.array([0.02, 0.025, -0.033, 0.01, 0.0008]).reshape((-1, 1))
        t = np.array([0.25, 1.0, 5.0, 20.0, 25.0]).reshape((-1, 1))
        ufr = 0.029
        alpha = 0.12

        t_target = np.array([0.25, 0.5, 1.0, 2.0, 2.5, 3.5, 5.0, 10.0, 20.0, 49.5, 125.0]).reshape((-1, 1))

        # Expected Output
        expected = np.array([0.02, 0.02417656, 0.025, 0.00361999, -0.00733027,
                            -0.02345319, -0.033, -0.01256218, 0.01, 0.00715949, 0.02015626]).reshape((-1, 1))

        # Actual Output
        actual = sw.fit_smithwilson_rates(rates_obs=r, t_obs=t, t_target=t_target, ufr=ufr, alpha=alpha)

        # Assert
        self.assertEqual(type(actual), type(expected), "Returned types not matching")
        self.assertTupleEqual(actual.shape, expected.shape, "Shapes not matching")
        np.testing.assert_almost_equal(actual, expected, decimal=8, err_msg="Fitted rates not matching")


import sys, os, pickle
from itertools import  product
from lifelib.projects.solvency2.scripts import solvency2
from lifelib.tests.data import round_signif

filepath = os.path.join(os.path.dirname(__file__), 'data_solvency2')

if '' not in sys.path:
    sys.path.insert(0, '')


def generate_data(model):
    data = []
    scr = model.SCR_life
    for t0, polid in product(range(0, 15, 5), (1, 101, 201)):
        data.append(round_signif(scr[t0, polid].SCR_life(), 10))
    return data


def save_date(data):
    with open(filepath, 'wb') as file:
        pickle.dump(data, file, protocol=4)


if __name__ == '__main__':
    save_date(generate_data(solvency2.build()))



"""Commutation functions and actuarial notations

The ``LifeTable`` Space includes Cells to calculate
commutation functions and actuarial notations.
``LifeTable`` is parameterized with
``Sex``, ``IntRate`` and ``TableID``. ``TableID`` and
``Sex`` are used in :py:func:`qx` below to identify
the mortality rates to be applied.

Example:

    .. code-block:: python

            >>> fastlife.LifeTable['M', 0.03, 3].AnnDuenx(x=30, n=10)
            8.752619688735953

            >>> fastlife.LifeTable['F', 0.03, 3].qx(x=50)
            0.00196

            >>> fastlife.LifeTable.MortalityTables()
                       1                 2                 3                 4
                       M        F        M        F        M        F        M        F
            0    0.00246  0.00210  0.00298  0.00252  0.00345  0.00298  0.00456  0.00383
            1    0.00037  0.00033  0.00045  0.00034  0.00051  0.00044  0.00069  0.00059
            2    0.00026  0.00023  0.00032  0.00025  0.00038  0.00030  0.00051  0.00041
            3    0.00018  0.00015  0.00022  0.00018  0.00027  0.00020  0.00037  0.00028
            4    0.00013  0.00011  0.00016  0.00013  0.00021  0.00014  0.00029  0.00021
            ..       ...      ...      ...      ...      ...      ...      ...      ...
            126  1.00000  1.00000  1.00000  1.00000  1.00000  1.00000  1.00000  1.00000
            127  1.00000  1.00000  1.00000  1.00000  1.00000  1.00000  1.00000  1.00000
            128  1.00000  1.00000  1.00000  1.00000  1.00000  1.00000  1.00000  1.00000
            129  1.00000  1.00000  1.00000  1.00000  1.00000  1.00000  1.00000  1.00000
            130  1.00000  1.00000  1.00000  1.00000  1.00000  1.00000  1.00000  1.00000

            [131 rows x 8 columns]

References:
    * `International actuarial notation by F.S.Perryman <https://www.casact.org/pubs/proceed/proceed49/49123.pdf>`_
    * `Actuarial notations on Wikipedia <https://en.wikipedia.org/wiki/Actuarial_notation>`_

.. rubric:: Space Parameters

Attributes:
    Sex: 'M' or 'F' to indicate male or female column in the mortality table.
    IntRate: Constant interest rate for discounting.
    TableID: ID of an ultimate mortality table by sex and age.

.. rubric:: References

Attributes:
    MortalityTables: `PandasData`_ object holding the data of mortality tables.
        The data is read from *MortalityTables.xlsx*. Defined also
        in :mod:`fastlife.model.LifeTable`,
        :mod:`fastlife.model.Input` and :mod:`fastlife.model.Projection.Assumptions`


.. _PandasData:
   https://docs.modelx.io/en/latest/reference/dataclient.html#pandasdata


"""

from modelx.serialize.jsonvalues import *

_formula = lambda Sex, IntRate, TableID: None

_bases = []

_allow_none = None

_spaces = []

# ---------------------------------------------------------------------------
# Cells

def AnnDuenx(x, n, k=1, f=0):
    """ The present value of an annuity-due.

    .. math::

        \\require{enclose}{}_{f|}\\ddot{a}_{x:\\enclose{actuarial}{n}}^{(k)}


    Args:
        x(int): age
        n(int): length of annuity payments in years
        k(int, optional): number of split payments in a year
        f(int, optional): waiting period in years

    """
    if Dx(x) == 0:
        return 0

    result = (Nx(x+f) - Nx(x+f+n)) / Dx(x)

    if k > 1:
        return result - (k-1) / (2*k) * (1 - Dx(x+f+n) / Dx(x))
    else:
        return result


def AnnDuex(x, k, f=0):
    """The present value of a lifetime annuity due.

    Args:
        x(int): age
        k(int, optional): number of split payments in a year
        f(int, optional): waiting period in years
    """
    if Dx(x) == 0:
        return 0

    result = (Nx(x+f)) / Dx(x)

    if k > 1:
        return result - (k-1) / (2*k)
    else:
        return result


def Ax(x, f=0):
    """The present value of a lifetime assurance on a person at age ``x``
    payable immediately upon death, optionally with an waiting period of ``f`` years.

    .. math::

        \\require{enclose}{}_{f|}\\overline{A}_{x}
    """
    if Dx(x) == 0:
        return 0
    else:
        return Mx(x+f) / Dx(x)


def Axn(x, n, f=0):
    """The present value of an assurance on a person at age ``x`` payable
    immediately upon death, optionally with an waiting period of ``f`` years.

    .. math::

        \\require{enclose}{}_{f|}\\overline{A}^{1}_{x:\\enclose{actuarial}{n}}

    """
    if Dx(x) == 0:
        return 0
    else:
        return (Mx(x+f) - Mx(x+f+n)) / Dx(x)


def Cx(x):
    """The commutation column :math:`\\overline{C_x}`.
    """

    return dx(x) * disc**(x+1/2)


def Dx(x):
    """The commutation column :math:`D_{x} = l_{x}v^{x}`.
    """
    return lx(x) * disc ** x


def Exn(x, n):
    """ The value of an endowment on a person at age ``x``
    payable after n years

    .. math::

        {}_{n}E_x

    """
    if Dx(x) == 0:
        return 0
    else:
        return Dx(x+n) / Dx(x)


def Mx(x):
    """The commutation column :math:`M_x`."""

    if x >= 110:
        return Dx(x)
    else:
        return Mx(x+1) + Cx(x)


def Nx(x):
    """The commutation column :math:`N_x`."""
    if x >= 110:    # TODO: Get the last age from the table
        return Dx(x)
    else:
        return Nx(x+1) + Dx(x)


def disc():
    """The discount factor :math:`v = 1/(1 + i)`."""
    return 1 / (1 + IntRate)


def dx(x):
    """The number of persons who die between ages ``x`` and ``x+1``"""
    return lx(x) * qx(x)


def lx(x):
    """The number of persons remaining at age ``x``. """
    if x == 0:
        return 100000
    else:
        return lx(x-1) - dx(x-1)


def qx(x):
    """Probability that a person at age ``x`` will die in one year."""
    return MortalityTables()[TableID, Sex][x]


# ---------------------------------------------------------------------------
# References

IntRate = 0.01

Sex = "M"

TableID = 1

MortalityTables = ("Pickle", 3020541952136)



from math import exp, log
import numpy as np

def calculate_prices(rates: np.ndarray, t: np.ndarray) -> np.ndarray:
    """Calculate prices from zero-coupon rates

    Args:
        rates: zero-coupon rates vector of length n
        t: time to maturity vector (in years) of length n

    Returns:
        Prices as vector of length n
    """

    return np.power(1 + rates, -t)


def ufr_discount_factor(ufr: float, t: np.ndarray) -> np.ndarray:
    """Calculate Ultimate Forward Rate (UFR) discount factors.

    Takes the UFR with a vector of maturities and returns for each of the
    maturities the discount factor
        d_UFR = e^(-UFR * t)

    Note that UFR is expected to be annualy compounded and that
    this function performs the calculation of the log return prior
    to applying the formula above.

    Args:
        ufr: Ultimate Forward Rate (annualized/annual compounding)
        t: time to maturity vector (in years) of length n

    Returns:
        UFR discount factors as vector of length n
    """

    # Convert annualized ultimate forward rate to log-return
    ufr = log(1 + ufr)
    return np.exp(-ufr * t)


def wilson_function(t1: np.ndarray, t2: np.ndarray, alpha: float, ufr: float) -> np.ndarray:
    """Calculate matrix of Wilson functions

    The Smith-Wilson method requires the calculation of a series of Wilson
    functions. The Wilson function is calculated for each maturity combination
    t1 and t2. If t1 and t2 are scalars, the result is a scalar. If t1 and t2
    are vectors of shape (m, 1) and (n, 1), then the result is a matrix of
    Wilson functions with shape (m, n) as defined on p. 16:
        W = e^(-UFR * (t1 + t2)) * (α * min(t1, t2) - 0.5 * e^(-α * max(t1, t2))
            * (e^(α * min(t1, t2)) - e^(-α * min(t1, t2))))

    Source: EIOPA QIS 5 Technical Paper; Risk-free interest rates – Extrapolation method; p.11ff
    https://wayback.archive-it.org/org-1495/20191229100044/https:/eiopa.europa.eu/Publications/QIS/ceiops-paper-extrapolation-risk-free-rates_en-20100802.pdf

    Args:
        t1: time to maturity vector of length m
        t2: time to maturity vector of length n
        alpha: Convergence speed parameter
        ufr: Ultimate Forward Rate (annualized/annual compounding)

    Returns:
        Wilson-matrix of shape (m, n) as numpy matrix
    """

    # Take time vectors of shape (nx1) and (mx1) and turn them into matrices of shape (mxn).
    # This is achieved by repeating the vectors along the axis 1. The operation is required
    # because the Wilson function needs all possible combinations of maturities to construct
    # the Wilson matrix
    m = len(t1)
    n = len(t2)
    t1_Mat = np.repeat(t1, n, axis=1)
    t2_Mat = np.repeat(t2, m, axis=1).T

    # Calculate the minimum and maximum of the two matrices
    min_t = np.minimum(t1_Mat, t2_Mat)
    max_t = np.maximum(t1_Mat, t2_Mat)

    # Calculate the UFR discount factor - p.16
    ufr_disc = ufr_discount_factor(ufr=ufr, t=(t1_Mat + t2_Mat))
    W = ufr_disc * (alpha * min_t - 0.5 * np.exp(-alpha * max_t) * \
        (np.exp(alpha * min_t) - np.exp(-alpha * min_t)))

    return W


def fit_parameters(rates: np.ndarray, t: np.ndarray, alpha: float, ufr: float) -> np.ndarray:
    """Calculate Smith-Wilson parameter vector ζ

    Given the Wilson-matrix, vector of discount factors and prices,
    the parameter vector can be calculated as follows (p.17):
        ζ = W^-1 * (μ - P)

    Source: EIOPA QIS 5 Technical Paper; Risk-free interest rates – Extrapolation method; p.11ff
    https://wayback.archive-it.org/org-1495/20191229100044/https:/eiopa.europa.eu/Publications/QIS/ceiops-paper-extrapolation-risk-free-rates_en-20100802.pdf

    Args:
        rates: Observed zero-coupon rates vector of length n
        t1: Observed time to maturity vector (in years) of length n
        alpha: Convergence speed parameter
        ufr: Ultimate Forward Rate (annualized/annual compounding)

    Returns:
        Wilson-matrix of shape (m, n) as numpy matrix
    """

    # Calcualte square matrix of Wilson functions, UFR discount vector and price vector
    # The price vector is calculated with zero-coupon rates and assumed face value of 1
    # For the estimation of zeta, t1 and t2 correspond both to the observed maturities
    W = wilson_function(t1=t, t2=t, alpha=alpha, ufr=ufr)
    mu = ufr_discount_factor(ufr=ufr, t=t)
    P = calculate_prices(rates=rates, t=t)

    # Calculate vector of parameters (p. 17)
    # To invert the Wilson-matrix, conversion to type matrix is required
    zeta = np.matrix(W).I * (mu - P)
    zeta = np.array(zeta) # Convert back to more general array type

    return zeta


def fit_smithwilson_rates(rates_obs: np.ndarray, t_obs: np.ndarray, t_target: np.ndarray,
                                alpha: float, ufr: float) -> np.ndarray:
    """Calculate zero-coupon yields with Smith-Wilson method based on observed rates.

    This function expects the rates and initial maturity vector to be
    before the Last Liquid Point (LLP). The targeted maturity vector can
    contain both, more granular maturity structure (interpolation) or terms after
    the LLP (extrapolation).

    The Smith-Wilson method calculated first the Wilson-matrix (p. 16):
        W = e^(-UFR * (t1 + t2)) * (α * min(t1, t2) - 0.5 * e^(-α * max(t1, t2))
            * (e^(α * min(t1, t2)) - e^(-α * min(t1, t2))))

    Given the Wilson-matrix, vector of discount factors and prices,
    the parameter vector can be calculated as follows (p.17):
        ζ = W^-1 * (μ - P)

    With the Smith-Wilson parameter and Wilson-matrix, the zero-coupon bond
    prices can be represented as (p. 18) in matrix notation:
        P = e^(-t * UFR) - W * zeta

    In the last case, t can be any maturity vector

    Source: EIOPA QIS 5 Technical Paper; Risk-free interest rates – Extrapolation method; p.11ff
    https://wayback.archive-it.org/org-1495/20191229100044/https:/eiopa.europa.eu/Publications/QIS/ceiops-paper-extrapolation-risk-free-rates_en-20100802.pdf

    Args:
        rates_obs: Initially observed zero-coupon rates vector before LLP of length n
        t_obs: Initially observed time to maturity vector (in years) of length n
        t_target: New targeted maturity vector (in years) with interpolated/extrapolated terms
        alpha: Convergence speed parameter
        ufr: Ultimate Forward Rate (annualized/annual compounding)

    Returns:
        Vector of zero-coupon rates with Smith-Wilson interpolated or extrapolated rates
    """

    # Convert list to numpy array and use reshape to convert from 1-d to 2-d array
    # E.g. reshape((-1, 1)) converts an input of shape (10,) with second dimension
    # being empty (1-d vector) to shape (10, 1) where second dimension is 1 (2-d vector)
    rates_obs = np.array(rates_obs).reshape((-1, 1))
    t_obs = np.array(t_obs).reshape((-1, 1))
    t_target = np.array(t_target).reshape((-1, 1))

    zeta = fit_parameters(rates=rates_obs, t=t_obs, alpha=alpha, ufr=ufr)
    ufr_disc = ufr_discount_factor(ufr=ufr, t=t_target)
    W = wilson_function(t1=t_target, t2=t_obs, alpha=alpha, ufr=ufr)

    # Price vector - equivalent to discounting with zero-coupon yields 1/(1 + r)^t
    # for prices where t_obs = t_target. All other matuirites are interpolated or extrapolated
    P = ufr_disc - W @ zeta  # '@' in numpy is the dot product of two matrices

    # Transform price vector to zero-coupon rate vector (1/P)^(1/t) - 1
    return np.power(1/P, 1/t_target) - 1



from modelx.serialize.jsonvalues import *

_formula = None

_bases = []

_allow_none = True

_spaces = []

# ---------------------------------------------------------------------------
# Cells

def AsmpLookup(asmp, prod=None, polt=None, gen=None):
    return  Assumption.get((asmp, prod, polt, gen), None)


def SpecLookup(spec, prod=None, polt=None, gen=None):
    return  ProductSpec.get((spec, prod, polt, gen), None)


def Factor(risk=None, shock=None, scope=None, extrakey=None):
    return FactorData[risk, shock, scope, extrakey]


# ---------------------------------------------------------------------------
# References

PolicyData = ("Pickle", 2325077945160)

MortalityTables = ("Pickle", 2325095460104)

AssumptionTables = ("Pickle", 2325101521608)

Scenarios = ("Pickle", 2325101521416)

DiscountRate = ("Pickle", 2325101684424)

PremWaiverCost = ("Pickle", 2325101684232)

Assumption = ("Pickle", 2325102163848)

ProductSpec = ("Pickle", 2325102164616)

CorrData = ("Pickle", 2325102165768)

FactorData = ("Pickle", 2325102235144)


"""Present values of cashflows

This Space is a child Space of :mod:`~ifrs17sim.model.OuterProj.InnerProj`,
and calculates present values of cashflows.

This Space is parameterized with ``t_rate``, which indicates that
the discount rates at time ``t_rate`` are used for discounting.

.. blockdiag::

   blockdiag {
     default_node_color="#D5E8D4";
     default_linecolor="#628E47";
     ifrs17sim [shape=roundedbox, linecolor="#7B99C5", color="#D4E8FC", width=96]
     ifrs17sim <- "OuterProj\\n[PolicyID, ScenID=1]" <- "InnerProj[t0]" [hstyle=composition];
     "OuterProj\\n[PolicyID, ScenID=1]" [stacked];
     "InnerProj[t0]" [stacked];
     "PV[t_rate]" [stacked];
     "InnerProj[t0]" <- "PV[t_rate]" [hstyle=composition]
   }

"""

from modelx.serialize.jsonvalues import *

def _formula(t_rate):
    refs = {'last_t': _space.parent.last_t,
            'InsurIF_Beg1': _space.parent.InsurIF_Beg1,
            'InsurIF_End': _space.parent.InsurIF_End,
            'PremIncome': _space.parent.PremIncome,
            'BenefitSurr': _space.parent.BenefitSurr,
            'BenefitDeath': _space.parent.BenefitDeath,
            'BenefitTotal': _space.parent.BenefitTotal,
            'ExpsCommTotal': _space.parent.ExpsCommTotal,
            'ExpsAcq': _space.parent.ExpsAcq,
            'ExpsMaint': _space.parent.ExpsMaint,
            'ExpsTotal': _space.parent.ExpsTotal,
            'DiscRate': _space.parent.parent[t_rate].DiscRate}

    return {'refs': refs}


_bases = []

_allow_none = None

_spaces = []

# ---------------------------------------------------------------------------
# Cells

def InterestNetCF(t):
    """Interest accreted on pv of net cashflows"""
    if t > last_t:
        return 0
    else:
        return (PV_NetCashflow(t)
                - PremIncome(t)
                + ExpsTotal(t)) * DiscRate(t)


def PV_BenefitDeath(t):
    """Present value of death benefits"""
    if t > last_t:
        return 0
    else:
        return (-BenefitDeath(t) + PV_BenefitDeath(t+1)) / (1 + DiscRate(t))


def PV_BenefitMat(t):
    """Present value of matuirty benefits"""
    if t > last_t:
        return 0
    else:
        return (-BenefitMat(t) + PV_BenefitMat(t+1)) / (1 + DiscRate(t))


def PV_BenefitSurr(t):
    """Present value of surrender benefits"""
    if t > last_t:
        return 0
    else:
        return (-BenefitSurr(t) + PV_BenefitSurr(t+1)) / (1 + DiscRate(t))


def PV_BenefitTotal(t):
    """Present value of total benefits"""
    if t > last_t:
        return 0
    else:
        return (-BenefitTotal(t) + PV_BenefitTotal(t+1)) / (1 + DiscRate(t))


def PV_Check(t):
    return PV_NetCashflow(t) - PV_NetCashflowForCheck(t)


def PV_ExpsAcq(t):
    """Present value of acquisition expenses"""
    if t > last_t:
        return 0
    else:
        return - ExpsAcq(t) + PV_ExpsAcq(t+1) / (1 + DiscRate(t))


def PV_ExpsCommTotal(t):
    """Present value of commission expenses"""
    if t > last_t:
        return 0
    else:
        return - ExpsCommTotal(t) + PV_ExpsCommTotal(t+1) / (1 + DiscRate(t))


def PV_ExpsMaint(t):
    """Present value of maintenance expenses"""
    if t > last_t:
        return 0
    else:
        return - ExpsMaint(t) + PV_ExpsMaint(t+1) / (1 + DiscRate(t))


def PV_ExpsTotal(t):
    """Present value of total expenses"""
    if t > last_t:
        return 0
    else:
        return - ExpsTotal(t) + PV_ExpsTotal(t+1) / (1 + DiscRate(t))


def PV_NetCashflow(t):
    """Present value of net cashflow"""
    return (PV_PremIncome(t)
            + PV_ExpsTotal(t)
            + PV_BenefitTotal(t))


def PV_NetCashflowForCheck(t):
    """Present value of net cashflow"""
    if t > last_t:
        return 0
    else:
        return (PremIncome(t)
                - ExpsTotal(t)
                - BenefitTotal(t) / (1 + DiscRate(t))
                + PV_NetCashflow(t+1) / (1 + DiscRate(t)))


def PV_PremIncome(t):
    """Present value of premium income"""
    if t > last_t:
        return 0
    else:
        return PremIncome(t) + PV_PremIncome(t+1) / (1 + DiscRate(t))


def PV_SumInsurIF(t):
    """Present value of insurance in-force"""
    if t > last_t:
        return 0
    else:
        return InsurIF_Beg1(t) + PV_SumInsurIF(t+1) / (1 + DiscRate(t))


# ---------------------------------------------------------------------------
# References

scen = ("Interface", ("....", "Economic"), "auto")


"""
Actual vs Estimated
===================

Lapse assumption changes based on previous year experience.
"""
import modelx as mx

model = mx.read_model("model")

# Policy point ID and aliases
polid = 171
outer = model.OuterProj[polid]
inner = outer.InnerProj

# Code block for overriding the default model

outer.asmp.SurrRateMult[1] = 2
outer.asmp.SurrRateMult[2] = 0.5
outer.asmp.SurrRateMult[3] = 1

inner[1].asmp.SurrRateMult[1] = 2
inner[2].asmp.SurrRateMult[2] = 0.5
inner[3].asmp.SurrRateMult[3] = 1

# Code block for drawing graphs

from draw_charts import draw_actest_pairs

draw_actest_pairs(outer, inner, ['PolsSurr', 'PolsIF_End'], 5, 10)




"""
:mod:`simplelife` cashflow
==========================

Draw a graph of liability cashflows of a simple whole life policy

*Run this sample now!* |binder simplelife_demo|

.. include:: /binderlinks.rst
   :start-after: Begin binder simplelife_demo
   :end-before: End binder simplelife_demo

"""
import pandas as pd
import modelx as mx
import seaborn as sns
sns.set()

polid = 171
proj = mx.read_model("model").Projection[polid]

# Draw NetCashflows Graph
data = {'NetCashflows': [proj.NetInsurCF[t] for t in range(50)]}
ax = pd.DataFrame(data).plot.line(marker='o', color='r')

# Draw componets of net cashflows

vars = ['PremIncome',
        'BenefitSurr',
        'BenefitDeath',
        'ExpsMaint',
        'ExpsCommTotal',
        'ExpsAcq']

df = proj.cells[vars].to_frame(range(50))

df[vars[1:]] = df[vars[1:]].mul(-1)   # Change outflows to negatives
df.plot(kind='bar', stacked=True, ax=ax, title='Insurance Cashflows')




"""
Account value to number of policies
=========================================

This graph uses the ``CashValue_ME_EX2`` model
from :doc:`/libraries/notebooks/savings/savings_example2`,
and shows path dependency of policy decrement.

The scatter plot shows how the number of policies at maturity varies
by the level of account value at maturity, and the graphs on the sides
show the distributions of the two factors.


.. seealso::

    * :doc:`/libraries/notebooks/savings/savings_example2` notebook in the :mod:`~savings` library

"""

# The code below is based on the sample script available on
# https://matplotlib.org/stable/gallery/lines_bars_and_markers/scatter_hist.html
# See accompanying LICENSE_MATPLOT.txt for its terms of conditions.

import numpy as np
import matplotlib.pyplot as plt
import modelx as mx

def scatter_hist(x, y, ax, ax_histx, ax_histy):
    # no labels
    ax_histx.tick_params(axis="x", labelbottom=False)
    ax_histy.tick_params(axis="y", labelleft=False)

    # the scatter plot:
    ax.scatter(x, y)

    def get_bins(data, binsize):
        ub, lb = np.max(data), np.min(data)
        binwidth = ((ub - lb) / binsize)
        return np.arange(lb, ub + binwidth, binwidth)

    xbins = get_bins(x, 100)
    ybins = get_bins(y, 100)
    ax_histx.hist(x, bins=xbins)
    ax_histy.hist(y, bins=ybins, orientation='horizontal')
    
    
# definitions for the axes
left, width = 0.1, 0.65
bottom, height = 0.1, 0.65
spacing = 0.005

rect_scatter = [left, bottom, width, height]
rect_histx = [left, bottom + height + spacing, width, 0.2]
rect_histy = [left + width + spacing, bottom, 0.2, height]

# start with a square Figure
fig = plt.figure(figsize=(8, 8))

ax = fig.add_axes(rect_scatter)
ax_histx = fig.add_axes(rect_histx, sharex=ax)
ax_histy = fig.add_axes(rect_histy, sharey=ax)
ax.grid(True), ax_histx.grid(True), ax_histy.grid(True)

m = mx.read_model('CashValue_ME_EX2')

x = m.Projection[5].claims_from_av(120, 'MATURITY')
y = m.Projection[5].pols_maturity(120)

# use the previously defined function
scatter_hist(x, y, ax, ax_histx, ax_histy)

m.close()
